# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import
import numpy as np
from scipy.optimize import brentq
from ..utils.misc import read_table, lookup_and_interpolate
from ..utils.math import bracket


def brown_shankland(depths):
    """
    Geotherm from :cite:`Brown1981`. NOTE: Valid only above 270 km

    :param depths: The list of depths at which to evaluate the geotherm.
        :math:`[m]`
    :type depths: list of floats

    :returns: The list of temperatures for each of the pressures. :math:`[K]`
    :rtype: list of floats
    """

    assert min(depths) >= min(table_brown_depth)
    assert max(depths) <= max(table_brown_depth)
    temperature = np.empty_like(depths)
    for i, depth in enumerate(depths):
        temperature[i] = lookup_and_interpolate(
            table_brown_depth, table_brown_temperature, depth
        )
    return temperature


def anderson(depths):
    """
    Geotherm from :cite:`anderson1982earth`.

    :param depths: The list of depths at which to evaluate the geotherm.
        :math:`[m]`
    :type depths: list of floats

    :returns: The list of temperatures for each of the pressures. :math:`[K]`
    :rtype: list of floats
    """
    assert min(depths) >= min(table_anderson_depth)
    assert max(depths) <= max(table_anderson_depth)
    temperature = np.empty_like(depths)
    for i, depth in enumerate(depths):
        temperature[i] = lookup_and_interpolate(
            table_anderson_depth, table_anderson_temperature, depth
        )
    return temperature


def stacey_continental(depths):
    """
    Continental geotherm from :cite:`stacey1977`.

    :param depths: The list of depths at which to evaluate the geotherm.
        :math:`[m]`
    :type depths: list of floats

    :returns: The list of temperatures for each of the pressures. :math:`[K]`
    :rtype: list of floats
    """
    assert min(depths) >= min(table_stacey_c_depth)
    assert max(depths) <= max(table_stacey_c_depth)
    temperature = np.empty_like(depths)
    for i, depth in enumerate(depths):
        temperature[i] = lookup_and_interpolate(
            table_stacey_c_depth, table_stacey_c_temperature, depth
        )
    return temperature


def stacey_oceanic(depths):
    """
    Oceanic geotherm from :cite:`stacey1977`.

    :param depths: The list of depths at which to evaluate the geotherm.
        :math:`[m]`
    :type depths: list of floats

    :returns: The list of temperatures for each of the pressures. :math:`[K]`
    :rtype: list of floats
    """
    assert min(depths) >= min(table_stacey_o_depth)
    assert max(depths) <= max(table_stacey_o_depth)
    temperature = np.empty_like(depths)
    for i, depth in enumerate(depths):
        temperature[i] = lookup_and_interpolate(
            table_stacey_o_depth, table_stacey_o_temperature, depth
        )
    return temperature


def adiabatic(pressures, T0, rock):
    """
    This calculates a geotherm based on an anchor temperature and a rock,
    assuming that the rock's temperature follows an adiabatic gradient with
    pressure. A good first guess is provided by integrating:

    .. math::
        \\frac{\\partial T}{\\partial P} = \\frac{ \\gamma  T}{ K_s }

    where :math:`\\gamma` is the Grueneisen parameter and :math:`K_s` is
    the adiabatic bulk modulus.

    :param pressures: The list of pressures in :math:`[Pa]` at which
        to evaluate the geotherm.
    :type pressures: list of floats

    :param T0: An anchor temperature, corresponding to the temperature of the first
        pressure in the list. :math:`[K]`
    :type T0: float

    :param rock: Composite for which we compute the adiabat.  From this material we
        must compute average Grueneisen parameters and adiabatic bulk moduli
        for each pressure/temperature.
    :type rock: :class:`burnman.composite`

    :returns: The list of temperatures for each pressure. :math:`[K]`
    :rtype: numpy.array of floats
    """

    rock.set_state(pressures[0], T0)
    S0 = rock.S

    delta_S = lambda T, P, rock, S0: S0 - rock.evaluate(["S"], [P], [T])[0][0]

    temperatures = np.empty_like(pressures)
    temperatures[0] = T0
    for i in range(1, len(pressures)):
        args = (pressures[i], rock, S0)
        sol = bracket(
            fn=delta_S,
            x0=(
                temperatures[i - 1]
                + (rock.gr * temperatures[i - 1] / rock.K_S)
                * (pressures[i] - pressures[i - 1])
            ),
            dx=1.0,
            args=args,
        )
        temperatures[i] = brentq(delta_S, sol[0], sol[1], args=args)

    return temperatures


table_brown = read_table("input_geotherm/brown_81.txt")
table_brown_depth = np.array(table_brown)[:, 0]
table_brown_temperature = np.array(table_brown)[:, 1]

table_anderson = read_table("input_geotherm/anderson_82.txt")
table_anderson_depth = np.array(table_anderson)[:, 0]
table_anderson_temperature = np.array(table_anderson)[:, 1]

table_stacey_c = read_table("input_geotherm/Stacey_1977_continents.txt")
table_stacey_c_depth = np.array(table_stacey_c)[:, 0]
table_stacey_c_temperature = np.array(table_stacey_c)[:, 1]

table_stacey_o = read_table("input_geotherm/Stacey_1977_oceans.txt")
table_stacey_o_depth = np.array(table_stacey_o)[:, 0]
table_stacey_o_temperature = np.array(table_stacey_o)[:, 1]



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.


# This is a standalone program that converts the Holland and Powell data format
# into the standard burnman format (printed to file)
# It only outputs properties of solid and melt endmembers
# - other endmembers are currently ignored.


import sys
import os.path
import pprint
from collections import OrderedDict

# hack to allow scripts to be placed in subdirectories next to burnman:
if not os.path.exists("burnman") and os.path.exists("../../../burnman"):
    sys.path.insert(1, os.path.abspath("../../.."))

from burnman.tools.chemistry import dictionarize_formula, formula_mass


if not os.path.isfile("tc-ds62.txt"):
    print("This code requires the data file tc-ds62.txt.")
    print(
        "This file is bundled with the software package THERMOCALC, "
        "which can be found here:"
    )
    print("http://www.metamorph.geo.uni-mainz.de/thermocalc/" "dataset6/index.html")
    print("")
    print("Please download the file and place it in this directory.")
    exit()

# Components
components = [
    "Si",
    "Ti",
    "Al",
    "Fe",
    "Mg",
    "Mn",
    "Ca",
    "Na",
    "K",
    "O",
    "H",
    "C",
    "Cl",
    "e-",
    "Ni",
    "Zr",
    "S",
    "Cu",
    "Cr",
]


class Endmember:
    def __init__(self, name, atoms, formula, sites, comp, H, S, V, Cp, a, k, flag, od):
        if flag != -1 and flag != -2 and k[0] > 0:
            formula = dictionarize_formula(formula)
            self.params = OrderedDict(
                [
                    ("name", name),
                    ("formula", formula),
                    ("equation_of_state", "hp_tmt"),
                    ("H_0", round(H * 1e3, 10)),
                    ("S_0", round(S * 1e3, 10)),
                    ("V_0", round(V * 1e-5, 15)),
                    (
                        "Cp",
                        [
                            round(Cp[0] * 1e3, 10),
                            round(Cp[1] * 1e3, 10),
                            round(Cp[2] * 1e3, 10),
                            round(Cp[3] * 1e3, 10),
                        ],
                    ),
                    ("a_0", a),
                    ("K_0", round(k[0] * 1e8, 10)),
                    ("Kprime_0", k[1]),
                    ("Kdprime_0", round(k[2] * 1e-8, 15)),
                    ("n", sum(formula.values())),
                    ("molar_mass", round(formula_mass(formula), 10)),
                ]
            )
        if flag == 1:
            self.landau_hp = OrderedDict(
                [
                    ("P_0", 1e5),
                    ("T_0", 298.15),
                    ("Tc_0", od[0]),
                    ("S_D", round(od[1] * 1e3, 10)),
                    ("V_D", round(od[2] * 1e-5, 10)),
                ]
            )

        if flag == 2:
            self.bragg_williams = OrderedDict(
                [
                    ("deltaH", round(od[0] * 1e3, 10)),
                    ("deltaV", round(od[1] * 1e-5, 15)),
                    ("Wh", round(od[2] * 1e3, 10)),
                    ("Wv", round(od[3] * 1e-5, 15)),
                    ("n", od[4]),
                    ("factor", od[5]),
                ]
            )

        if flag == -2:
            formula = dictionarize_formula(formula)
            self.params = OrderedDict(
                [
                    ("name", name),
                    ("formula", formula),
                    ("equation_of_state", "hp_tmtL"),
                    ("H_0", round(H * 1e3, 10)),
                    ("S_0", round(S * 1e3, 10)),
                    ("V_0", round(V * 1e-5, 15)),
                    (
                        "Cp",
                        [
                            round(Cp[0] * 1e3, 10),
                            round(Cp[1] * 1e3, 10),
                            round(Cp[2] * 1e3, 10),
                            round(Cp[3] * 1e3, 10),
                        ],
                    ),
                    ("a_0", a),
                    ("K_0", round(k[0] * 1e8, 10)),
                    ("Kprime_0", k[1]),
                    ("Kdprime_0", round(k[2] * 1e-8, 15)),
                    ("dKdT_0", round(k[3] * 1e8, 15)),
                    ("n", sum(formula.values())),
                    ("molar_mass", round(formula_mass(formula), 10)),
                ]
            )


# Read dataset
with open("tc-ds62.txt", "r") as f:
    ds = [line.split() for line in f]


def getmbr(ds, mbr):
    for i in range(0, int(ds[0][0])):
        if ds[i * 4 + 3][0] == mbr:
            atoms = 0.0
            formula = ""
            for j in range(3, len(ds[i * 4 + 3]) - 1, 2):
                atoms += float(ds[i * 4 + 3][j])
                formula = (
                    formula
                    + components[int(ds[i * 4 + 3][j - 1]) - 1]
                    + str(round(float(ds[i * 4 + 3][j]), 10))
                )

            sites = int(ds[i * 4 + 3][1])
            comp = list(map(float, ds[i * 4 + 3][2 : (len(ds[i * 4 + 3]) - 1)]))
            H = float(ds[i * 4 + 4][0])
            S = float(ds[i * 4 + 4][1])
            V = float(ds[i * 4 + 4][2])
            Cp = list(map(float, ds[i * 4 + 5]))
            a = float(ds[i * 4 + 6][0])
            k = list(map(float, ds[i * 4 + 6][1:4]))
            od = list(map(float, ds[i * 4 + 6][5:]))

            if mbr.endswith("L"):
                flag = -2
                od = [0]
                k.append(float(ds[i * 4 + 6][4]))
            else:
                flag = int(ds[i * 4 + 6][4])

            endmember = Endmember(
                mbr, atoms, formula, sites, comp, H, S, V, Cp, a, k, flag, od
            )
            return endmember


with open("HP_2011_ds62.py", "w") as outfile:
    outfile.write(
        "# This file is part of BurnMan - a thermoelastic and "
        "thermodynamic toolkit\n# for the Earth and Planetary "
        "Sciences\n"
        "# Copyright (C) 2012 - 2021 by the BurnMan team, "
        "released under the GNU\n# GPL v2 or later.\n\n\n"
        '"""\n'
        "HP_2011_ds62\n"
        "^^^^^^^^^^^^\n"
        "\n"
        "Endmember minerals from Holland and Powell 2011 and "
        "references therein.\n"
        "Update to dataset version 6.2.\n"
        "The values in this document are all in S.I. units,\n"
        "unlike those in the original tc-ds62.txt.\n"
        "File autogenerated using HPdata_to_burnman.py.\n"
        '"""\n'
        "\n\n"
        "from ..mineral import Mineral\n\n"
    )

    outfile.write('"""\n' "ENDMEMBERS\n" '"""\n\n')

    def pprint_ordered_dict(d, leading_string, extra_whitespace=0, width=200):
        old_leading_string = "OrderedDict["
        whitespace = " " * (
            len(leading_string) + extra_whitespace - len(old_leading_string)
        )
        s = pprint.pformat(d, width=width)
        s = s.replace("),\n", f",\n{whitespace}")
        s = s.replace("), ", f",\n{whitespace}")
        s = s.replace("', ", "': ").replace("(", "")
        s = s.replace(old_leading_string, leading_string + "{").replace(")])", "}")
        return s

    formula = "0"
    for i in range(int(ds[0][0])):
        mbr = ds[i * 4 + 3][0]
        M = getmbr(ds, mbr)
        if mbr == "and":  # change silly abbreviation
            mbr = "andalusite"

        # Print parameters
        if hasattr(M, "params"):
            outfile.write("\n")
            outfile.write(
                "class {0} (Mineral):\n".format(mbr) + "    def __init__(self):\n"
            )

            s = pprint_ordered_dict(M.params, leading_string="        self.params = ")
            s = s.replace("000000.0", "e6")
            outfile.write(s)
            outfile.write("\n")

            # Print property modifiers (if they exist)
            if hasattr(M, "landau_hp") or hasattr(M, "bragg_williams"):
                outfile.write("        self.property_modifiers = [[")
                if hasattr(M, "landau_hp"):
                    s = pprint_ordered_dict(
                        M.landau_hp, leading_string="'landau_hp', ", extra_whitespace=49
                    )
                    outfile.write(s)

                if hasattr(M, "bragg_williams"):
                    s = pprint_ordered_dict(
                        M.bragg_williams,
                        leading_string="'bragg_williams', ",
                        extra_whitespace=49,
                    )
                    outfile.write(s)

                outfile.write("]]\n")

            outfile.write("        Mineral.__init__(self)\n\n")

    outfile.write(
        "\ndef cov():\n"
        '    """\n'
        "    A function which loads and returns the "
        "variance-covariance matrix of the\n"
        "    zero-point energies of all the endmembers "
        "in the dataset.\n\n"
        "    Returns\n"
        "    -------\n"
        "    cov : dictionary\n"
        "        Dictionary keys are:\n"
        "        - endmember_names: a list of endmember names, and\n"
        "        - covariance_matrix: a 2D variance-covariance "
        "array for the\n"
        "          endmember zero-point energies of formation\n"
        '    """\n\n'
        "    from .HP_2011_ds62_cov import cov\n"
        "    return cov\n"
    )

# Process uncertainties
with open("HP_2011_ds62_cov.py", "w") as outfile:
    outfile.write(
        "# This file is part of BurnMan - a thermoelastic and "
        "thermodynamic toolkit\n# for the Earth and Planetary "
        "Sciences\n"
        "# Copyright (C) 2012 - 2021 by the BurnMan team, "
        "released under the GNU\n# GPL v2 or later.\n\n\n"
        '"""\n'
        "HP_2011 (ds-62) zero-point energy covariance matrix\n"
        "Derived from Holland and Powell 2011 and references "
        "therein\n"
        "Update to dataset version 6.2\n"
        "The values in this document are all in S.I. units,\n"
        "unlike those in the original tc-ds62.txt\n"
        "File autogenerated using HPdata_to_burnman.py\n"
        '"""\n\n'
        "from numpy import array\n\n"
        "cov = "
    )

    import numpy as np

    n_mbrs = int(ds[0][0])

    names = []
    for i in range(n_mbrs):
        names.append(ds[i * 4 + 3][0])

    cov = []
    for i in range(n_mbrs * 4 + 4, len(ds) - 2):
        cov.extend(map(float, ds[i]))

    i_utr = np.triu_indices(n_mbrs)
    i_ltr = np.tril_indices(n_mbrs)
    M = np.zeros((n_mbrs, n_mbrs))

    M[i_utr] = cov[1:]
    M[i_ltr] = M.T[i_ltr]

    M = M * 1.0e6  # (kJ/mol)^2 -> (J/mol)^2

    d = {"endmember_names": names, "covariance_matrix": M}

    np.set_printoptions(threshold=sys.maxsize)

    pp = pprint.PrettyPrinter(indent=0, width=160, depth=3, stream=outfile)
    pp.pprint(d)

    outfile.write("\n")



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


"""
Other minerals
^^^^^^^^^^^^^^

"""
from __future__ import absolute_import

from ..classes import mineral_helpers as helpers
from ..classes.mineral import Mineral
from ..utils.chemistry import dictionarize_formula, formula_mass

from .SLB_2011 import periclase, wuestite, mg_perovskite, fe_perovskite


class liquid_iron(Mineral):
    """
    Liquid iron equation of state from Anderson and Ahrens (1994)
    """

    def __init__(self):
        formula = "Fe1.0"
        formula = dictionarize_formula(formula)
        m = formula_mass(formula)
        rho_0 = 7019.0
        V_0 = m / rho_0
        D = 7766.0
        Lambda = 1146.0
        self.params = {
            "name": "liquid iron",
            "formula": formula,
            "equation_of_state": "aa",
            "P_0": 1.0e5,
            "T_0": 1811.0,
            "S_0": 100.0,  # to fit
            "molar_mass": m,
            "V_0": V_0,
            "E_0": 0.0,
            "K_S": 109.7e9,
            "Kprime_S": 4.661,
            "Kprime_prime_S": -0.043e-9,
            "grueneisen_0": 1.735,
            "grueneisen_prime": -0.130 / m * 1.0e-6,
            "grueneisen_n": -1.870,
            "a": [248.92 * m, 289.48 * m],
            "b": [0.4057 * m, -1.1499 * m],
            "Theta": [1747.3, 1.537],
            "theta": 5000.0,
            "lmda": [-325.23 * m, 302.07 * m, 30.45 * m],
            "xi_0": 282.67 * m,
            "F": [D / rho_0, Lambda / rho_0],
            "n": sum(formula.values()),
            "molar_mass": m,
        }
        Mineral.__init__(self)


class ZSB_2013_mg_perovskite(Mineral):
    def __init__(self):
        self.params = {
            "equation_of_state": "slb3",
            "V_0": 24.45e-6,
            "K_0": 250.5e9,
            "Kprime_0": 4.01,
            "G_0": 172.9e9,
            "Gprime_0": 1.74,
            "molar_mass": 0.1000,
            "n": 5,
            "Debye_0": 905.9,
            "grueneisen_0": 1.44,
            "q_0": 1.09,
            "eta_s_0": 2.13,
        }  # 2.6

        self.uncertainties = {
            "err_K_0": 3.0e9,
            "err_Kprime_0": 0.1,
            "err_G_0": 2.0e9,
            "err_Gprime_0": 0.0,
            "err_Debye_0": 5.0,
            "err_grueneisen_0": 0.05,
            "err_q_0": 0.3,
            "err_eta_s_0": 0.3,
        }

        Mineral.__init__(self)


class ZSB_2013_fe_perovskite(Mineral):
    def __init__(self):
        self.params = {
            "equation_of_state": "slb3",
            "V_0": 25.49e-6,
            "K_0": 272.0e9,
            "Kprime_0": 4.1,
            "G_0": 133.0e9,
            "Gprime_0": 1.4,
            "molar_mass": 0.1319,
            "n": 5,
            "Debye_0": 871.0,
            "grueneisen_0": 1.57,
            "q_0": 1.1,
            "eta_s_0": 2.3,
        }  # 2.3

        self.uncertainties = {
            "err_K_0": 40e9,
            "err_Kprime_0": 1.0,
            "err_G_0": 40e9,
            "err_Gprime_0": 0.0,
            "err_Debye_0": 26.0,
            "err_grueneisen_0": 0.3,
            "err_q_0": 1.0,
            "err_eta_s_0": 1.0,
        }

        Mineral.__init__(self)


class Speziale_fe_periclase(helpers.HelperSpinTransition):
    def __init__(self):
        helpers.HelperSpinTransition.__init__(
            self, 60.0e9, Speziale_fe_periclase_LS(), Speziale_fe_periclase_HS()
        )
        self.cite = "Speziale et al. 2007"


class Speziale_fe_periclase_HS(Mineral):
    """
    Speziale et al. 2007, Mg#=83
    """

    def __init__(self):
        self.params = {
            "equation_of_state": "mgd3",
            "V_0": 22.9e-6,
            "K_0": 157.5e9,
            "Kprime_0": 3.92,
            "molar_mass": 0.04567,
            "n": 2,
            "Debye_0": 587,
            "grueneisen_0": 1.46,
            "q_0": 1.2,
        }
        Mineral.__init__(self)


class Speziale_fe_periclase_LS(Mineral):
    """
    Speziale et al. 2007, Mg#=83
    """

    def __init__(self):
        self.params = {
            "equation_of_state": "mgd3",
            "V_0": 21.49e-6,
            "K_0": 186.0e9,
            "Kprime_0": 4.6,
            "molar_mass": 0.04567,
            "n": 2,
            "Debye_0": 587.0,
            "grueneisen_0": 1.46,
            "q_0": 1.2,
        }
        Mineral.__init__(self)


class Liquid_Fe_Anderson(Mineral):
    """
    Anderson & Ahrens, 1994 JGR
    """

    def __init__(self):
        self.params = {
            "equation_of_state": "bm4",
            "V_0": 7.95626e-6,
            "K_0": 109.7e9,
            "Kprime_0": 4.66,
            "Kprime_prime_0": -0.043e-9,
            "molar_mass": 0.055845,
        }
        Mineral.__init__(self)


class Fe_Dewaele(Mineral):
    """
    Dewaele et al., 2006, Physical Review Letters
    """

    def __init__(self):
        self.params = {
            "equation_of_state": "vinet",
            "V_0": 6.75e-6,
            "K_0": 163.4e9,
            "Kprime_0": 5.38,
            "molar_mass": 0.055845,
            "n": 1,
        }
        Mineral.__init__(self)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit
# for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2022 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import

import importlib
import warnings
import numpy as np
from ..utils.chemistry import process_solution_chemistry
from .solutionmodel import _ideal_activities_fct
from .solutionmodel import _non_ideal_hessian_fct, _non_ideal_interactions_fct
from .solutionmodel import _non_ideal_hessian_subreg
from .solutionmodel import _non_ideal_interactions_subreg
from .solutionmodel import logish, inverseish
from .. import constants

try:
    ag = importlib.import_module("autograd")
except ImportError as err:
    print(
        f"Warning: {err}. "
        "For full functionality of BurnMan, please install autograd."
    )


class ElasticSolutionModel(object):
    """
    This is the base class for an Elastic solution model, intended for use
    in defining solutions and performing thermodynamic calculations
    on them.  All minerals of type :class:`burnman.Solution` use
    a solution model for defining how the endmembers in the solution
    interact.

    A user wanting a new solution model should define the functions included
    in the base class. All of the functions in the base class return zero,
    so if the user-defined solution model does not implement them,
    they essentially have no effect, and the Helmholtz energy and
    pressure of a solution will be equal to the weighted arithmetic
    averages of the different endmember values.
    """

    def __init__(self):
        """
        Does nothing.
        """
        pass

    def excess_helmholtz_energy(self, volume, temperature, molar_fractions):
        """
        Given a list of molar fractions of different phases,
        compute the excess Helmholtz free energy of the solution.
        The base class implementation assumes that the excess Helmholtz
        energy is zero.

        :param volume: Volume at which to evaluate the solution model. [m^3/mol]
        :type volume: float

        :param temperature: Temperature at which to evaluate the solution. [K]
        :type temperature: float

        :param molar_fractions: List of molar fractions of the different
            endmembers in solution.
        :type molar_fractions: list of floats

        :returns: The excess Helmholtz energy.
        :rtype: float
        """
        return np.dot(
            np.array(molar_fractions),
            self.excess_partial_helmholtz_energies(
                volume, temperature, molar_fractions
            ),
        )

    def excess_pressure(self, volume, temperature, molar_fractions):
        """
        Given a list of molar fractions of different phases,
        compute the excess pressure of the solution.
        The base class implementation assumes that the excess pressure is zero.

        :param volume: Volume at which to evaluate the solution model. [m^3/mol]
        :type volume: float

        :param temperature: Temperature at which to evaluate the solution. [K]
        :type temperature: float

        :param molar_fractions: List of molar fractions of the different
            endmembers in solution.
        :type molar_fractions: list of floats

        :returns: The excess pressure of the solution.
        :rtype: float
        """
        return np.dot(
            molar_fractions,
            self.excess_partial_pressures(volume, temperature, molar_fractions),
        )

    def excess_entropy(self, volume, temperature, molar_fractions):
        """
        Given a list of molar fractions of different phases,
        compute the excess entropy of the solution.
        The base class implementation assumes that the excess entropy is zero.

        :param volume: Volume at which to evaluate the solution model. [m^3/mol]
        :type volume: float

        :param temperature: Temperature at which to evaluate the solution. [K]
        :type temperature: float

        :param molar_fractions: List of molar fractions of the different
            endmembers in solution.
        :type molar_fractions: list of floats

        :returns: The excess entropy of the solution.
        :rtype: float
        """
        return np.dot(
            molar_fractions,
            self.excess_partial_entropies(volume, temperature, molar_fractions),
        )

    def excess_enthalpy(self, volume, temperature, molar_fractions):
        """
        Given a list of molar fractions of different phases,
        compute the excess enthalpy of the solution.
        The base class implementation assumes that the excess enthalpy is zero.

        :param volume: Volume at which to evaluate the solution model. [m^3/mol]
        :type volume: float

        :param temperature: Temperature at which to evaluate the solution. [K]
        :type temperature: float

        :param molar_fractions: List of molar fractions of the different
            endmembers in solution.
        :type molar_fractions: list of floats

        :returns: The excess enthalpy of the solution.
        :rtype: float
        """
        return (
            self.excess_helmholtz_energy(volume, temperature, molar_fractions)
            + temperature * self.excess_entropy(volume, temperature, molar_fractions)
            - volume * self.excess_pressure(volume, temperature, molar_fractions)
        )

    def excess_partial_helmholtz_energies(self, volume, temperature, molar_fractions):
        """
        Given a list of molar fractions of different phases,
        compute the excess Helmholtz energy for each endmember of the solution.
        The base class implementation assumes that the excess Helmholtz energy
        is zero.

        :param volume: Volume at which to evaluate the solution model. [m^3/mol]
        :type volume: float

        :param temperature: Temperature at which to evaluate the solution. [K]
        :type temperature: float

        :param molar_fractions: List of molar fractions of the different
            endmembers in solution.
        :type molar_fractions: list of floats

        :returns: The excess Helmholtz energy of each endmember
        :rtype: numpy.array
        """
        return np.zeros_like(molar_fractions)

    def excess_partial_entropies(self, volume, temperature, molar_fractions):
        """
        Given a list of molar fractions of different phases,
        compute the excess entropy for each endmember of the solution.
        The base class implementation assumes that the excess entropy
        is zero (true for mechanical solutions).

        :param volume: Volume at which to evaluate the solution model. [m^3/mol]
        :type volume: float

        :param temperature: Temperature at which to evaluate the solution. [K]
        :type temperature: float

        :param molar_fractions: List of molar fractions of the different
            endmembers in solution.
        :type molar_fractions: list of floats

        :returns: The excess entropy of each endmember.
        :rtype: numpy.array
        """
        return np.zeros_like(molar_fractions)

    def excess_partial_pressures(self, volume, temperature, molar_fractions):
        """
        Given a list of molar fractions of different phases,
        compute the excess pressure for each endmember of the solution.
        The base class implementation assumes that the excess pressure
        is zero.

        :param volume: Volume at which to evaluate the solution model. [m^3/mol]
        :type volume: float

        :param temperature: Temperature at which to evaluate the solution. [K]
        :type temperature: float

        :param molar_fractions: List of molar fractions of the different
            endmembers in solution.
        :type molar_fractions: list of floats

        :returns: The excess pressure of each endmember.
        :rtype: numpy.array
        """
        return np.zeros_like(np.array(molar_fractions))


class ElasticMechanicalSolution(ElasticSolutionModel):
    """
    An extremely simple class representing a mechanical solution model.
    A mechanical solution experiences no interaction between endmembers.
    Therefore, unlike ideal solutions there is no entropy of mixing;
    the total Helmholtz energy of the solution is equal to the
    dot product of the molar Helmholtz energies and molar fractions
    of the constituent materials.
    """

    def __init__(self, endmembers):
        self.endmembers = endmembers
        self.n_endmembers = len(endmembers)
        self.formulas = [e[1] for e in endmembers]

    def excess_helmholtz_energy(self, volume, temperature, molar_fractions):
        return 0.0

    def excess_pressure(self, volume, temperature, molar_fractions):
        return 0.0

    def excess_entropy(self, volume, temperature, molar_fractions):
        return 0.0

    def excess_partial_helmholtz_energies(self, volume, temperature, molar_fractions):
        return np.zeros_like(molar_fractions)

    def excess_partial_pressures(self, volume, temperature, molar_fractions):
        return np.zeros_like(molar_fractions)

    def excess_partial_entropies(self, volume, temperature, molar_fractions):
        return np.zeros_like(molar_fractions)


class ElasticIdealSolution(ElasticSolutionModel):
    """
    A class representing an ideal solution model.
    Calculates the excess Helmholtz energy and entropy due to configurational
    entropy. Excess internal energy and volume are equal to zero.

    The multiplicity of each type of site in the structure is allowed to
    change linearly as a function of endmember proportions. This class
    is therefore equivalent to the entropic part of
    a Temkin-type model :cite:`Temkin1945`.
    """

    def __init__(self, endmembers):
        self.endmembers = endmembers
        self.n_endmembers = len(endmembers)
        self.formulas = [e[1] for e in endmembers]

        # Process solution chemistry
        process_solution_chemistry(self)

        self._calculate_endmember_configurational_entropies()

    def _calculate_endmember_configurational_entropies(self):
        S_conf = -(
            constants.gas_constant
            * (self.endmember_noccupancies * logish(self.endmember_occupancies)).sum(-1)
        )
        self.endmember_configurational_entropies = S_conf

    def excess_partial_helmholtz_energies(self, volume, temperature, molar_fractions):
        return self._ideal_excess_partial_helmholtz(temperature, molar_fractions)

    def excess_partial_entropies(self, volume, temperature, molar_fractions):
        return self._ideal_excess_partial_entropies(temperature, molar_fractions)

    def excess_partial_pressures(self, volume, temperature, molar_fractions):
        return np.zeros((self.n_endmembers))

    def helmholtz_hessian(self, volume, temperature, molar_fractions):
        hess_S = self._ideal_entropy_hessian(temperature, molar_fractions)
        return -temperature * hess_S

    def entropy_hessian(self, volume, temperature, molar_fractions):
        hess_S = self._ideal_entropy_hessian(temperature, molar_fractions)
        return hess_S

    def pressure_hessian(self, volume, temperature, molar_fractions):
        return np.zeros((len(molar_fractions), len(molar_fractions)))

    def _configurational_entropy(self, molar_fractions):
        site_noccupancies = np.einsum(
            "i, ij", molar_fractions, self.endmember_noccupancies
        )
        site_multiplicities = np.einsum(
            "i, ij", molar_fractions, self.site_multiplicities
        )
        site_occupancies = site_noccupancies * inverseish(site_multiplicities)
        conf_entropy = -(
            constants.gas_constant
            * (site_noccupancies * logish(site_occupancies)).sum(-1)
        )
        return conf_entropy

    def _ideal_excess_partial_helmholtz(self, temperature, molar_fractions):
        return -(
            temperature
            * self._ideal_excess_partial_entropies(temperature, molar_fractions)
        )

    def _ideal_excess_partial_entropies(self, temperature, molar_fractions):
        return -(constants.gas_constant * self._log_ideal_activities(molar_fractions))

    def _ideal_entropy_hessian(self, temperature, molar_fractions):
        hessian = -constants.gas_constant * self._log_ideal_activity_derivatives(
            molar_fractions
        )
        return hessian

    def _log_ideal_activities(self, molar_fractions):
        site_noccupancies = np.einsum(
            "i, ij", molar_fractions, self.endmember_noccupancies
        )
        site_multiplicities = np.einsum(
            "i, ij", molar_fractions, self.site_multiplicities
        )

        lna = np.einsum(
            "ij, j->i",
            self.endmember_noccupancies,
            logish(site_noccupancies) - logish(site_multiplicities),
        )

        normalisation_constants = (
            self.endmember_configurational_entropies / constants.gas_constant
        )
        return lna + normalisation_constants

    def _log_ideal_activity_derivatives(self, molar_fractions):
        site_noccupancies = np.einsum(
            "i, ij", molar_fractions, self.endmember_noccupancies
        )
        site_multiplicities = np.einsum(
            "i, ij", molar_fractions, self.site_multiplicities
        )

        dlnadp = np.einsum(
            "pj, qj, j->pq",
            self.endmember_noccupancies,
            self.endmember_noccupancies,
            inverseish(site_noccupancies),
        ) - np.einsum(
            "pj, qj, j->pq",
            self.endmember_noccupancies,
            self.site_multiplicities,
            inverseish(site_multiplicities),
        )

        return dlnadp

    def _ideal_activities(self, molar_fractions):
        return _ideal_activities_fct(
            molar_fractions,
            self.endmember_noccupancies,
            self.n_endmembers,
            self.n_occupancies,
            self.site_multiplicities,
            self.endmember_configurational_entropies,
        )


class ElasticAsymmetricRegularSolution(ElasticIdealSolution):
    """
    Solution model implementing the asymmetric regular solution model
    formulation as described in :cite:`HP2003`.

    The excess nonconfigurational Helmholtz energy is given by the
    expression:

    .. math::
        \\mathcal{F}_{\\textrm{excess}} = \\alpha^T p (\\phi^T W \\phi)

    :math:`\\alpha` is a vector of van Laar parameters governing asymmetry
    in the excess properties.

    .. math::
        \\phi_i = \\frac{\\alpha_i p_i}{\\sum_{k=1}^{n} \\alpha_k p_k},
        W_{ij} = \\frac{2 w_{ij}}{\\alpha_i + \\alpha_j} \\textrm{for i<j}
    """

    def __init__(
        self,
        endmembers,
        alphas,
        energy_interaction,
        pressure_interaction=None,
        entropy_interaction=None,
    ):
        self.n_endmembers = len(endmembers)

        # Create array of van Laar parameters
        self.alphas = np.array(alphas)

        # Create 2D arrays of interaction parameters
        self.We = np.triu(2.0 / (self.alphas[:, np.newaxis] + self.alphas), 1)
        self.We[np.triu_indices(self.n_endmembers, 1)] *= np.array(
            [i for row in energy_interaction for i in row]
        )

        if entropy_interaction is not None:
            self.Ws = np.triu(2.0 / (self.alphas[:, np.newaxis] + self.alphas), 1)
            self.Ws[np.triu_indices(self.n_endmembers, 1)] *= np.array(
                [i for row in entropy_interaction for i in row]
            )
        else:
            self.Ws = np.zeros((self.n_endmembers, self.n_endmembers))

        if pressure_interaction is not None:
            self.Wp = np.triu(2.0 / (self.alphas[:, np.newaxis] + self.alphas), 1)
            self.Wp[np.triu_indices(self.n_endmembers, 1)] *= np.array(
                [i for row in pressure_interaction for i in row]
            )
        else:
            self.Wp = np.zeros((self.n_endmembers, self.n_endmembers))

        # initialize ideal solution model
        ElasticIdealSolution.__init__(self, endmembers)

    def _phi(self, molar_fractions):
        phi = self.alphas * molar_fractions
        phi = np.divide(phi, np.sum(phi))
        return phi

    def _non_ideal_interactions(self, W, molar_fractions):
        # -sum(sum(qi.qj.Wij*)
        # equation (2) of Holland and Powell 2003
        phi = self._phi(molar_fractions)
        return _non_ideal_interactions_fct(
            phi, np.array(molar_fractions), self.n_endmembers, self.alphas, W
        )

    def _non_ideal_excess_partial_helmholtz(self, volume, temperature, molar_fractions):
        Eint = self._non_ideal_interactions(self.We, molar_fractions)
        Sint = self._non_ideal_interactions(self.Ws, molar_fractions)
        Pint = self._non_ideal_interactions(self.Wp, molar_fractions)
        return Eint - temperature * Sint - volume * Pint

    def excess_partial_helmholtz_energies(self, volume, temperature, molar_fractions):
        ideal_helmholtz = ElasticIdealSolution._ideal_excess_partial_helmholtz(
            self, temperature, molar_fractions
        )
        non_ideal_helmholtz = self._non_ideal_excess_partial_helmholtz(
            volume, temperature, molar_fractions
        )
        return ideal_helmholtz + non_ideal_helmholtz

    def excess_partial_entropies(self, volume, temperature, molar_fractions):
        ideal_entropies = ElasticIdealSolution._ideal_excess_partial_entropies(
            self, temperature, molar_fractions
        )
        non_ideal_entropies = self._non_ideal_interactions(self.Ws, molar_fractions)
        return ideal_entropies + non_ideal_entropies

    def excess_partial_pressures(self, volume, temperature, molar_fractions):
        return self._non_ideal_interactions(self.Wp, molar_fractions)

    def helmholtz_hessian(self, volume, temperature, molar_fractions):
        ideal_entropy_hessian = ElasticIdealSolution._ideal_entropy_hessian(
            self, temperature, molar_fractions
        )
        phi = self._phi(molar_fractions)
        nonideal_helmholtz_hessian = _non_ideal_hessian_fct(
            phi,
            molar_fractions,
            self.n_endmembers,
            self.alphas,
            self.We - temperature * self.Ws - volume * self.Wp,
        )

        return nonideal_helmholtz_hessian - temperature * ideal_entropy_hessian

    def entropy_hessian(self, volume, temperature, molar_fractions):
        ideal_entropy_hessian = ElasticIdealSolution._ideal_entropy_hessian(
            self, temperature, molar_fractions
        )
        phi = self._phi(molar_fractions)
        nonideal_entropy_hessian = _non_ideal_hessian_fct(
            phi, molar_fractions, self.n_endmembers, self.alphas, self.Ws
        )
        return ideal_entropy_hessian + nonideal_entropy_hessian

    def pressure_hessian(self, volume, temperature, molar_fractions):
        phi = self._phi(molar_fractions)
        return _non_ideal_hessian_fct(
            phi, molar_fractions, self.n_endmembers, self.alphas, self.Wp
        )


class ElasticSymmetricRegularSolution(ElasticAsymmetricRegularSolution):
    """
    Solution model implementing the symmetric regular solution model.
    This is a special case of the
    :class:`burnman.solutionmodel.AsymmetricRegularSolution` class.
    """

    def __init__(
        self,
        endmembers,
        energy_interaction,
        pressure_interaction=None,
        entropy_interaction=None,
    ):
        alphas = np.ones(len(endmembers))
        ElasticAsymmetricRegularSolution.__init__(
            self,
            endmembers,
            alphas,
            energy_interaction,
            pressure_interaction,
            entropy_interaction,
        )


class ElasticSubregularSolution(ElasticIdealSolution):
    """
    Solution model implementing the subregular solution model formulation
    as described in :cite:`HW1989`. The excess conconfigurational
    Helmholtz energy is given by the expression:

    .. math::
        \\mathcal{F}_{\\textrm{excess}} = \\sum_i \\sum_{j > i} (p_i p_j^2
        W_{ij} + p_j p_i^2 W_{ji} + \\sum_{k > j > i} p_i p_j p_k W_{ijk})

    Interaction parameters are inserted into a 3D interaction matrix during
    initialization to make use of numpy vector algebra.

    :param endmembers: A list of all the independent endmembers in the solution.
        The first item of each list gives the Mineral object corresponding
        to the endmember. The second item gives the site-species formula.
    :type endmembers: list of lists
    :param energy_interaction: The binary endmember interaction energies.
        Each interaction[i, j-i-1, 0] corresponds to W(i,j), while
        interaction[i, j-i-1, 1] corresponds to W(j,i).
    :type energy_interaction: list of list of lists
    :param pressure_interaction: The binary endmember interaction pressures.
        Each interaction[i, j-i-1, 0] corresponds to W(i,j), while
        interaction[i, j-i-1, 1] corresponds to W(j,i).
    :type pressure_interaction: list of list of lists
    :param entropy_interaction: The binary endmember interaction entropies.
        Each interaction[i, j-i-1, 0] corresponds to W(i,j), while
        interaction[i, j-i-1, 1] corresponds to W(j,i).
    :type entropy_interaction: list of list of lists
    :param energy_ternary_terms: The ternary interaction energies.
        Each list should contain four entries: the indices i, j, k
        and the value of the interaction.
    :type energy_ternary_terms: list of lists
    :param pressure_ternary_terms: The ternary interaction pressures.
        Each list should contain four entries: the indices i, j, k
        and the value of the interaction.
    :type pressure_ternary_terms: list of lists
    :param entropy_ternary_terms: The ternary interaction entropies.
        Each list should contain four entries:
        the indices i, j, k and the value of the interaction.
    :type entropy_ternary_terms: list of lists
    """

    def __init__(
        self,
        endmembers,
        energy_interaction,
        pressure_interaction=None,
        entropy_interaction=None,
        energy_ternary_terms=None,
        pressure_ternary_terms=None,
        entropy_ternary_terms=None,
    ):
        """
        Initialization function for the SubregularSolution class.
        """

        self.n_endmembers = len(endmembers)

        # Create 3D arrays of interaction parameters
        self.Wijke = np.zeros(
            shape=(self.n_endmembers, self.n_endmembers, self.n_endmembers)
        )
        self.Wijks = np.zeros_like(self.Wijke)
        self.Wijkp = np.zeros_like(self.Wijke)

        # setup excess enthalpy interaction matrix
        for i in range(self.n_endmembers):
            for j in range(i + 1, self.n_endmembers):
                w0 = energy_interaction[i][j - i - 1][0] / 2.0
                w1 = energy_interaction[i][j - i - 1][1] / 2.0
                self.Wijke[:, i, j] += w0
                self.Wijke[:, j, i] += w1

                self.Wijke[i, j, j] += w0
                self.Wijke[j, i, i] += w1

                self.Wijke[i, j, i] -= w0
                self.Wijke[j, i, j] -= w1

        if energy_ternary_terms is not None:
            for i, j, k, v in energy_ternary_terms:
                self.Wijke[i, j, k] += v

        if entropy_interaction is not None:
            for i in range(self.n_endmembers):
                for j in range(i + 1, self.n_endmembers):
                    w0 = entropy_interaction[i][j - i - 1][0] / 2.0
                    w1 = entropy_interaction[i][j - i - 1][1] / 2.0
                    self.Wijks[:, i, j] += w0
                    self.Wijks[:, j, i] += w1

                    self.Wijks[i, j, j] += w0
                    self.Wijks[j, i, i] += w1

                    self.Wijks[i, j, i] -= w0
                    self.Wijks[j, i, j] -= w1

        if entropy_ternary_terms is not None:
            for i, j, k, v in entropy_ternary_terms:
                self.Wijks[i, j, k] += v

        if pressure_interaction is not None:
            for i in range(self.n_endmembers):
                for j in range(i + 1, self.n_endmembers):
                    w0 = pressure_interaction[i][j - i - 1][0] / 2.0
                    w1 = pressure_interaction[i][j - i - 1][1] / 2.0
                    self.Wijkp[:, i, j] += w0
                    self.Wijkp[:, j, i] += w1

                    self.Wijkp[i, j, j] += w0
                    self.Wijkp[j, i, i] += w1

                    self.Wijkp[i, j, i] -= w0
                    self.Wijkp[j, i, j] -= w1

        if pressure_ternary_terms is not None:
            for i, j, k, v in pressure_ternary_terms:
                self.Wijkv[i, j, k] += v

        # initialize ideal solution model
        ElasticIdealSolution.__init__(self, endmembers)

    def _non_ideal_function(self, Wijk, molar_fractions):
        n = len(molar_fractions)
        return _non_ideal_interactions_subreg(molar_fractions, n, Wijk)

    def _non_ideal_interactions(self, molar_fractions):
        # equation (6') of Helffrich and Wood, 1989
        Eint = self._non_ideal_function(self.Wijke, molar_fractions)
        Sint = self._non_ideal_function(self.Wijks, molar_fractions)
        Pint = self._non_ideal_function(self.Wijkp, molar_fractions)
        return Eint, Sint, Pint

    def _non_ideal_excess_partial_helmholtz(self, volume, temperature, molar_fractions):
        Eint, Sint, Pint = self._non_ideal_interactions(molar_fractions)
        return Eint - temperature * Sint - volume * Pint

    def excess_partial_helmholtz_energies(self, volume, temperature, molar_fractions):
        ideal_helmholtz = ElasticIdealSolution._ideal_excess_partial_helmholtz(
            self, temperature, molar_fractions
        )
        non_ideal_helmholtz = self._non_ideal_excess_partial_helmholtz(
            volume, temperature, molar_fractions
        )
        return ideal_helmholtz + non_ideal_helmholtz

    def excess_partial_entropies(self, volume, temperature, molar_fractions):
        ideal_entropies = ElasticIdealSolution._ideal_excess_partial_entropies(
            self, temperature, molar_fractions
        )
        non_ideal_entropies = self._non_ideal_function(self.Wijks, molar_fractions)
        return ideal_entropies + non_ideal_entropies

    def excess_partial_pressures(self, volume, temperature, molar_fractions):
        non_ideal_pressures = self._non_ideal_function(self.Wijkp, molar_fractions)
        return non_ideal_pressures

    def helmholtz_hessian(self, volume, temperature, molar_fractions):
        n = len(molar_fractions)
        ideal_entropy_hessian = ElasticIdealSolution._ideal_entropy_hessian(
            self, temperature, molar_fractions
        )
        nonideal_helmholtz_hessian = _non_ideal_hessian_subreg(
            molar_fractions,
            n,
            self.Wijke - temperature * self.Wijks - volume * self.Wijkp,
        )

        return nonideal_helmholtz_hessian - temperature * ideal_entropy_hessian

    def entropy_hessian(self, volume, temperature, molar_fractions):
        n = len(molar_fractions)
        ideal_entropy_hessian = ElasticIdealSolution._ideal_entropy_hessian(
            self, temperature, molar_fractions
        )
        nonideal_entropy_hessian = _non_ideal_hessian_subreg(
            molar_fractions, n, self.Wijks
        )
        return ideal_entropy_hessian + nonideal_entropy_hessian

    def pressure_hessian(self, volume, temperature, molar_fractions):
        n = len(molar_fractions)
        return _non_ideal_hessian_subreg(molar_fractions, n, self.Wijkp)


class ElasticFunctionSolution(ElasticIdealSolution):
    """
    Solution model implementing a generalized elastic solution model.
    The extensive excess nonconfigurational Helmholtz energy is
    provided as a function by the user.

    Derivatives are calculated using the autograd module,
    and so the user-defined excess Helmholtz energy function
    should be defined using autograd-friendly expressions.

    :param endmembers: A list of all the independent endmembers in the solution.
        The first item of each list gives the Mineral object corresponding
        to the endmember. The second item gives the site-species formula.
    :type endmembers: list of lists

    :param excess_helmholtz_function: The nonconfigurational
        Helmholtz energy function with arguments volume,
        temperature and molar_amounts, in that order.
        Note that the function must be extensive; if the molar amounts
        are doubled, the Helmholtz energy must also double.
    :type excess_helmholtz_function: function
    """

    def __init__(self, endmembers, excess_helmholtz_function):
        """
        Initialization function for the GeneralSolution class.
        """

        # initialize ideal solution model
        ElasticIdealSolution.__init__(self, endmembers)

        self.n_endmembers = len(endmembers)
        self._excess_helmholtz_function = excess_helmholtz_function

        self._non_ideal_excess_partial_helmholtz = ag.jacobian(
            excess_helmholtz_function, argnum=2
        )

        def partial_entropies(volume, temperature, molar_amounts):
            with warnings.catch_warnings(record=True):
                warnings.simplefilter("always")
                return -ag.jacobian(self._non_ideal_excess_partial_helmholtz, argnum=1)(
                    volume, temperature, molar_amounts
                )

        self._non_ideal_excess_partial_entropies = partial_entropies

        def partial_pressures(volume, temperature, molar_amounts):
            with warnings.catch_warnings(record=True):
                warnings.simplefilter("always")
                return -ag.jacobian(self._non_ideal_excess_partial_helmholtz, argnum=0)(
                    volume, temperature, molar_amounts
                )

        self.excess_partial_pressures = partial_pressures

        self._non_ideal_helmholtz_hessian = ag.jacobian(
            self._non_ideal_excess_partial_helmholtz, argnum=2
        )

        def entropy_hess(volume, temperature, molar_amounts):
            with warnings.catch_warnings(record=True):
                warnings.simplefilter("always")
                return ag.jacobian(partial_entropies, argnum=2)(
                    volume, temperature, molar_amounts
                )

        self._non_ideal_entropy_hessian = entropy_hess

        def pressure_hess(volume, temperature, molar_amounts):
            with warnings.catch_warnings(record=True):
                warnings.simplefilter("always")
                return ag.jacobian(partial_pressures, argnum=2)(
                    volume, temperature, molar_amounts
                )

        self.pressure_hessian = pressure_hess

    def excess_partial_helmholtz_energies(self, volume, temperature, molar_fractions):
        ideal_helmholtz = ElasticIdealSolution._ideal_excess_partial_helmholtz(
            self, temperature, molar_fractions
        )
        non_ideal_helmholtz = self._non_ideal_excess_partial_helmholtz(
            volume, temperature, molar_fractions
        )
        return ideal_helmholtz + non_ideal_helmholtz

    def excess_partial_entropies(self, volume, temperature, molar_fractions):
        ideal_entropies = ElasticIdealSolution._ideal_excess_partial_entropies(
            self, temperature, molar_fractions
        )
        non_ideal_entropies = self._non_ideal_excess_partial_entropies(
            volume, temperature, molar_fractions
        )
        return ideal_entropies + non_ideal_entropies

    def helmholtz_hessian(self, volume, temperature, molar_fractions):
        ideal_entropy_hessian = ElasticIdealSolution._ideal_entropy_hessian(
            self, temperature, molar_fractions
        )
        nonideal_helmholtz_hessian = self._non_ideal_helmholtz_hessian(
            volume, temperature, molar_fractions
        )

        return nonideal_helmholtz_hessian - temperature * ideal_entropy_hessian

    def entropy_hessian(self, volume, temperature, molar_fractions):
        ideal_entropy_hessian = ElasticIdealSolution._ideal_entropy_hessian(
            self, temperature, molar_fractions
        )
        nonideal_entropy_hessian = self._non_ideal_entropy_hessian(
            volume, temperature, molar_fractions
        )
        return ideal_entropy_hessian + nonideal_entropy_hessian



import numpy as np
from scipy.linalg import lu_factor, lu_solve
from collections import namedtuple


def solve_constraint_lagrangian(x, jac_x, c_x, c_prime):
    """
    Function which solves the problem
    minimize || J.dot(x_mod - x) ||
    subject to C(x_mod) = 0
    via the method of Lagrange multipliers.

    :param x: Parameter values at x.
    :type x: 1D numpy array

    :param jac_x: The (estimated, approximate or exact) value of the Jacobian J(x).
    :type jac_x: 2D numpy array.

    :param c_x: Values of the constraints at x.
    :type c_x: 1D numpy array

    :param c_prime: The Jacobian of the constraints (A, where A.x + b = 0).
    :type c_prime: 2D array of floats

    :returns: An array containing the parameter values which minimize the L2-norm
        of any function which has the Jacobian jac_x, and another array containing
        the multipliers for each of the equality constraints.
    :rtype: tuple
    """
    n_x = len(x)
    n = n_x + len(c_x)
    A = np.zeros((n, n))
    b = np.zeros(n)

    JTJ = jac_x.T.dot(jac_x)
    A[:n_x, :n_x] = JTJ / np.linalg.norm(JTJ) * n * n  # includes scaling
    A[:n_x, n_x:] = c_prime.T
    A[n_x:, :n_x] = c_prime
    b[n_x:] = c_x

    luA = lu_factor(A)
    dx_m = lu_solve(luA, -b)  # lu_solve computes the solution of ax = b

    x_mod = x + dx_m[:n_x]
    lagrange_multipliers = dx_m[n_x:]
    return (x_mod, lagrange_multipliers)


def damped_newton_solve(
    F,
    J,
    guess,
    tol=1.0e-6,
    max_iterations=100,
    lambda_bounds=lambda dx, x: (1.0e-8, 1.0),
    linear_constraints=(0.0, np.array([-1.0])),
    store_iterates=False,
):
    """
    Solver for the multivariate nonlinear system F(x)=0
    with Jacobian J(x), using the damped affine invariant modification
    to Newton's method (Deuflhard, 1974;1975;2004).
    Here we follow the algorithm as described in Nowak and Weimann (1991):
    [Technical Report TR-91-10, Algorithm B], modified to accept
    linear inequality constraints.

    Linear inequality constraints are provided by the arrays constraints_A and
    constraints_b. The constraints are satisfied if A*x + b <= 0.
    If any constraints are not satisfied by the current
    value of lambda, lambda is reduced to satisfy all the constraints.

    If a current iterate starting point (x_i) lies on one or more constraints
    and the Newton step violates one or more of those constraints, then
    the next step is calculated via the method of Lagrangian multipliers,
    minimizing the L2-norm of F(x_i+1) subject to the violated constraints.

    Successful termination of the solver is based on three criteria:
        - all(np.abs(dx (simplified newton step) < tol))
        - all(np.abs(dx (full Newton step) < sqrt(10*tol))) (avoids pathology) and
        - lambda = lambda_bounds(dx, x)[1] (lambda = 1 for a full Newton step).

    If these criteria are not satisfied, iterations continue until one of the following
    occurs:
        - the value of lmda is reduced to its minimum value (for v. nonlinear problems)
        - successive iterations have descent vectors which violate the constraints
        - the maximum number of iterations (given by max_iterations) is reached.

    Information on the root (or lack of root) obtained by the solver is provided
    in the returned namedtuple.

    :param F: Function returning the system function F(x) as a 1D numpy array.
    :type F: function of x

    :param J: Function returning the Jacobian function J(x) as a 2D numpy array.
    :type J: function of x

    :param guess: Starting guess for the solver.
    :type guess: 1D numpy.array

    :param tol: Tolerance(s) for termination.
    :type tol: float or array of floats

    :param max_iterations: Maximum number of iterations for the solver.
    :type max_iterations: int

    :param lambda_bounds: A function of dx and x that returns
        a tuple of floats corresponding to the minimum and maximum
        allowed fractions of the full newton step (dx).
    :type lambda bounds: function of dx and x

    :param linear_constraints: tuple of a 2D numpy array (A) and 1D numpy array (b)
        Constraints are satisfied if A.x + b < eps
    :type linear_constraints: tuple of a 2D numpy.array (A) and 1D numpy.array (b)

    :returns: A namedtuple with the following attributes:

        - x: The solution vector [1D numpy array].
        - F: The evaluated function F(x) [1D numpy array].
        - F_norm: Euclidean norm of F(x) [float].
        - J: The evaluated Jacobian J(x) [2D numpy array].
        - n_it: Number of iterations [int].
        - code: Numerical description of the solver termination [int].
            - 0: Successful convergence
            - 1: Failure due to solver hitting lower lambda bound
            - 2: Failure due to descent vector crossing constraints
            - 3: Failure due to solver reaching maximum number of iterations
        - text: Description of the solver termination [str].
        - success: Solver convergence state [bool].
        - iterates: [namedtuple]
            Only present if store_iterates=True
            Includes the following attributes:
                - x: list of 1D numpy arrays of floats
                    The parameters for each iteration
                - F: list of 2D numpy arrays of floats
                    The function for each iteration
                - lmda: list of floats
                    The value of the damping parameter for each iteration
    :rtype: namedtuple
    """

    # Make sure damping factor is within bounds, and that the bounds are reasonable
    # Problem classes in Nowak and Weimann (1991); [lmda_min, lmda_max]:
    # linear: [0.1, 1.]
    # mildly nonlinear: [1.e-4, 1.]
    # highly nonlinear: [1.e-2, 1.e-4]
    # extremely nonlinear: [1.e-4, 1.e-8]
    eps = 2.0 * np.finfo(float).eps

    def update_lmda(x, dx, h, lmda_bounds):
        assert (
            lmda_bounds[1] < 1.0 + eps
        ), "The highest upper bound for lambda is 1. (a full Newton step)"
        assert (
            lmda_bounds[0] > 1.0e-8 - eps
        ), "The lowest lower bound for lambda is 1.e-8 (suitable only for extremely nonlinear systems)"

        lmda_j = min(1.0 / (h + eps), lmda_bounds[1])  # this is lmda_j^0
        return max(lmda_j, lmda_bounds[0])

    def constraints(x):
        return np.dot(linear_constraints[0], x) + linear_constraints[1]

    assert np.all(
        constraints(guess) < eps
    ), "The starting guess is outside the supplied constraints."

    if not isinstance(tol, float):
        assert len(tol) < len(
            guess
        ), "tol must either be a float or an array like guess."

    sol = namedtuple(
        "Solution", ["x", "n_it", "F", "F_norm", "J", "code", "text", "success"]
    )

    # evaluate system
    sol.x = guess
    sol.F = F(sol.x)

    if store_iterates:
        sol.iterates = namedtuple("iterates", ["x", "F", "lmda"])
        sol.iterates.x = [sol.x]
        sol.iterates.F = [sol.F]
        sol.iterates.lmda = [0.0]

    # Begin Newton loop

    # Some dummy variables for the first h calculation (h = 0)
    lmda = 0.0
    dxprev = [1.0]
    dxbar = [1.0]

    sol.n_it = 0
    n_constraints = len(constraints(sol.x))
    minimum_lmda = False
    converged = False
    persistent_bound_violation = False
    while (
        sol.n_it < max_iterations
        and not minimum_lmda
        and not persistent_bound_violation
        and not converged
    ):
        sol.J = J(sol.x)  # evaluate Jacobian
        luJ = lu_factor(sol.J)  # storing the factorisation saves time later
        dx = lu_solve(luJ, -sol.F)  # compute ordinary Newton step
        dx_norm = np.linalg.norm(dx, ord=2)
        lmda_bounds = lambda_bounds(dx, sol.x)
        h = (
            lmda
            * np.linalg.norm((dxbar - dx), ord=2)
            * dx_norm
            / (np.linalg.norm(dxprev, ord=2) * np.linalg.norm(dxbar, ord=2))
        )
        lmda = update_lmda(sol.x, dx, h, lmda_bounds)

        # Create the (k+1)^0 values
        x_j = sol.x + lmda * dx

        # Check that all constraints are satisfied. If not, adjust lambda.
        # This must be done just before every call to F() *if* lambda
        # has been increased:
        c_x_j = constraints(x_j)
        if not np.all(
            c_x_j < eps
        ):  # x allowed to lie on constraints but not in forbidden area
            c_x = constraints(sol.x)
            violated_constraints = sorted(
                [
                    (i, c_x[i] / (c_x[i] - c_x_j[i]))
                    for i in range(n_constraints)
                    if c_x_j[i] >= eps
                ],
                key=lambda x: x[1],
            )
            lmda = lmda * violated_constraints[0][1]
            x_j = sol.x + lmda * dx

        # If the same current iterate is on a constraint,
        # and a very small lambda causes the next iterate to leave the
        # feasible region, then a new step direction must be found,
        # along with a new guess for lmda
        # We do this here using Lagrange multipliers
        if lmda < eps:
            active_constraint_indices = [
                i for i, vc in violated_constraints if vc < eps
            ]
            inactive_constraint_indices = [
                i for i, vc in violated_constraints if vc >= eps
            ]
            c_newton = constraints(sol.x + dx)[active_constraint_indices]
            c_A = linear_constraints[0][active_constraint_indices]
            x_n = sol.x + dx  # newton iterate
            if np.linalg.matrix_rank(c_A) == len(
                dx
            ):  # if true, we must leave a constraint here
                n_act = len(active_constraint_indices)
                for i_rm in range(n_act):
                    potential_active_indices = [
                        active_constraint_indices[i] for i in range(n_act) if i != i_rm
                    ]
                    c_newton = constraints(sol.x + dx)[potential_active_indices]
                    c_A = linear_constraints[0][potential_active_indices]
                    x_m = solve_constraint_lagrangian(x_n, sol.J, c_newton, c_A)[0]
                    if constraints(x_m)[active_constraint_indices[i_rm]] < 0.0:
                        break
            else:
                x_m = solve_constraint_lagrangian(x_n, sol.J, c_newton, c_A)[0]

            dx = x_m - sol.x
            lmda_bounds = lambda_bounds(dx, sol.x)
            lmda = lmda_bounds[1]  # no a-priori maximum limit
            x_j = sol.x + lmda * dx

            # Check that the solution is still able to converge, i.e.
            # that the constraints aren't stopping our approach to a potential root
            x_j_min = (
                sol.x + lmda_bounds[0] * dx
            )  # because lmda must be getting smaller, no need to check constraints
            F_j_min = F(x_j_min)
            dxbar_j_min = lu_solve(luJ, -F_j_min)
            dxbar_j_min_norm = np.linalg.norm(dxbar_j_min, ord=2)

            # Newton step size must be decreasing and dx must be non-zero
            if dxbar_j_min_norm > dx_norm or np.linalg.norm(dx, ord=2) < eps:
                persistent_bound_violation = True

            # Now we need to check for newly violated constraints
            n_inactive = len(inactive_constraint_indices)
            c_x_j = constraints(x_j)[inactive_constraint_indices]
            if not np.all(
                c_x_j < eps
            ):  # x allowed to lie on constraints but not in forbidden area
                c_x = constraints(sol.x)[inactive_constraint_indices]
                violated_constraints = sorted(
                    [
                        (i, c_x[i] / (c_x[i] - c_x_j[i]))
                        for i in range(n_inactive)
                        if c_x_j[i] >= eps
                    ],
                    key=lambda x: x[1],
                )
                lmda = lmda * violated_constraints[0][1]
                x_j = sol.x + lmda * dx

        F_j = F(x_j)
        dxbar_j = lu_solve(luJ, -F_j)  # this is the simplified newton step
        dxbar_j_norm = np.linalg.norm(dxbar_j, ord=2)

        if (
            all(np.abs(dxbar_j) < tol)  # <- Success requirements
            and all(np.abs(dx) < np.sqrt(10.0 * tol))  # <- avoids pathological cases
            and np.abs(lmda - lmda_bounds[1]) < eps
        ):  # <- end on a maximal newton step
            require_posteriori_loop = False  # <- No need for the a posteriori loop
            converged = True  # <- Successful convergence
        else:
            require_posteriori_loop = True

        # Begin the a posteriori loop
        while (
            require_posteriori_loop
            and not minimum_lmda
            and not persistent_bound_violation
        ):
            # Monotonicity check
            # always based on the Newton step, even if on a constraint
            if dxbar_j_norm <= dx_norm:
                if (
                    dxbar_j_norm < eps
                ):  # <- occasionally the simplified newton step finds the exact solution
                    converged = True
                dxbar = dxbar_j
                sol.x = x_j
                sol.F = F_j

                require_posteriori_loop = False  # return to Newton step
                sol.n_it += 1  # move to next iteration
                dxprev = dx  # to calculate the next value of h
            else:
                if np.abs(lmda - lmda_bounds[0]) < eps:
                    minimum_lmda = True
                h_j = (
                    (2.0 / lmda)
                    * np.linalg.norm((dxbar_j - (1.0 - lmda) * dx), ord=2)
                    / dx_norm
                )
                lmda_j = min(lmda_bounds[1], 1.0 / h_j)
                lmda = min(lmda_j, lmda / 2.0)
                lmda = max(
                    lmda, lmda_bounds[0]
                )  # allows a check of monotonicity once at minimum lmda

                x_j = (
                    sol.x + lmda * dx
                )  # because lmda must be getting smaller, no need to check constraints
                F_j = F(x_j)
                dxbar_j = lu_solve(luJ, -F_j)
                dxbar_j_norm = np.linalg.norm(dxbar_j, ord=2)

        if store_iterates:
            sol.iterates.x.append(sol.x)
            sol.iterates.F.append(sol.F)
            sol.iterates.lmda.append(lmda)

    if converged and not persistent_bound_violation:
        sol.x = x_j + dxbar_j
        # Even if the solver succeeds, there may be a small chance
        # that the last simplified Newton step
        # shifts the solution just outside the constraints.
        # If so, shift the solution back to the allowed region
        c_x = constraints(sol.x)
        if not np.all(
            c_x <= 0.0
        ):  # x allowed to lie on constraints but not in forbidden area
            sol.x -= dxbar_j

    sol.F = F(sol.x)
    sol.F_norm = np.linalg.norm(sol.F, ord=2)
    sol.J = J(sol.x)

    if store_iterates:
        sol.iterates.x = np.array(sol.iterates.x)
        sol.iterates.F = np.array(sol.iterates.F)

    sol.success = False
    if converged:
        sol.success = True
        sol.code = 0
        sol.text = "The solver successfully found a root after {0} iterations".format(
            sol.n_it
        )
    elif minimum_lmda:
        sol.code = 1
        sol.text = "The function is too non-linear for lower lambda bound ({0})".format(
            lmda_bounds[0]
        )
    elif persistent_bound_violation:
        sol.code = 2
        sol.text = "The descent vector crosses the constraints with the following indices: {0}".format(
            [i for i, lmda in violated_constraints]
        )
    elif sol.n_it == max_iterations:
        sol.code = 3
        sol.text = "The solver reached max_iterations ({0})".format(max_iterations)
    else:
        raise Exception("Unknown termination of solver")
    return sol



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2024 by the BurnMan team, released under the GNU
# GPL v2 or later.


# This is a script that converts a HPx-eos data format into the standard burnman
# format (printed to stdout). For example, the dataset file
# tc-thermoinput-igneous-2022-01-23/tc-ig50NCKFMASHTOCr.txt
# can be processed  via the commands
# python hpx_eos_to_burnman.py > ../../minerals/ig50NCKFMASHTOCr.py; black ../../minerals/ig50NCKFMASHTOCr.py
# the text "from . import ig50NCKFMASHTOCr must then be added to the minerals __init__.py

import numpy as np
import logging
import sys
from sympy import Symbol, prod, sympify
from sympy.parsing.sympy_parser import parse_expr
from burnman.constants import gas_constant
from burnman.minerals import HGP_2018_ds633, HP_2011_ds62
from datetime import date

ds = [
    ["tc-thermoinput-igneous-2022-01-23/tc-ig50NCKFMASHTOCr.txt", "HGP_2018_ds633"],
    ["tc-thermoinput-igneous-2022-01-23/tc-ig50NCKFMASTOCr.txt", "HGP_2018_ds633"],
    ["tc-thermoinput-metabasite-2022-01-30/tc-mb50NCKFMASHTO.txt", "HP_2011_ds62"],
    ["tc-thermoinput-metapelite-2022-01-23/tc-mp50KFMASH.txt", "HP_2011_ds62"],
    ["tc-thermoinput-metapelite-2022-01-23/tc-mp50MnNCKFMASHTO.txt", "HP_2011_ds62"],
    ["tc-thermoinput-metapelite-2022-01-23/tc-mp50NCKFMASHTO.txt", "HP_2011_ds62"],
]

ignore_solutions = ["liq", "L", "fl"]  # currently unreadable Temkin models

# logging.basicConfig(stream=sys.stderr, level=logging.DEBUG)
logging.basicConfig(stream=sys.stderr, level=logging.ERROR)
current_year = str(date.today())[:4]


def reverse_polish(lines):
    # get symbols
    syms = []
    for line in lines:
        n_terms = int(line[0])
        i = 2
        for term in range(n_terms):
            n_add = int(line[i])
            syms.extend(line[i + 2 : i + 2 * n_add + 2][::2])
            i += 2 * n_add + 2
    unique_syms = list(np.unique(sorted(syms)))
    sympy_syms = {u: Symbol(u) for u in unique_syms}

    # get expressions
    expr = 0.0
    for line in lines:
        n_terms = int(line[0])
        i = 2
        terms = []
        for term in range(n_terms):
            n_add = int(line[i])
            en = parse_expr(line[i - 1])  # this is the constant
            for j in range(n_add):
                en += parse_expr(line[i + 1 + 2 * j]) * sympy_syms[line[i + 2 + 2 * j]]
            terms.append(en)
            # syms.extend(l[i+2:i+2*n_add + 2][::2])
            # print("a", n_add)
            i += 2 * n_add + 2
        product = prod(terms)
        expr += product
    return expr


def ordering_modifier(property_modifier, ordered):
    mtype, params = property_modifier
    if mtype == "bragg_williams":
        if ordered:
            return np.array([0.0, 0.0, 0.0])
        else:
            n = params["n"]
            if params["factor"] > 0.0:
                f = [params["factor"], params["factor"]]
            else:
                f = [1.0, -params["factor"]]

            S_disord = (
                -gas_constant
                * (
                    f[0] * (np.log(1.0 / (n + 1.0)) + n * np.log(n / (n + 1.0)))
                    + f[1]
                    * (n * np.log(1.0 / (n + 1.0)) + n * n * np.log(n / (n + 1.0)))
                )
                / (n + 1.0)
            )
            return np.array([params["deltaH"], S_disord, params["deltaV"]])
    else:
        if params["T_0"] < params["Tc_0"]:
            Q_0 = np.power((params["Tc_0"] - params["T_0"]) / params["Tc_0"], 0.25)
        else:
            Q_0 = 0.0

        if ordered:
            E_ord = (
                params["Tc_0"] * params["S_D"] * (Q_0 * Q_0 - np.power(Q_0, 6.0) / 3.0)
                - (
                    params["S_D"] * params["Tc_0"] * 2.0 / 3.0
                    - params["V_D"] * params["P_0"]
                )
                - params["P_0"] * params["V_D"] * Q_0 * Q_0
            )
            S_ord = params["S_D"] * (Q_0 * Q_0 - 1.0)
            V_ord = params["V_D"] * (Q_0 * Q_0 - 1.0)
            return np.array([E_ord, S_ord, V_ord])

        else:
            E_disord = (
                params["Tc_0"] * params["S_D"] * (Q_0 * Q_0 - np.power(Q_0, 6.0) / 3.0)
                - params["P_0"] * params["V_D"] * Q_0 * Q_0
            )
            S_disord = params["S_D"] * Q_0 * Q_0
            V_disord = params["V_D"] * Q_0 * Q_0
            return np.array([E_disord, S_disord, V_disord])


for solution_file, mbr_dataset in ds:
    logging.debug(f"{solution_file}, {mbr_dataset}")
    dataset = eval(mbr_dataset)
    out_ss = ""
    out_make = ""

    i = 0
    data = []
    with open(solution_file, "r") as file:
        # Read each line in the file one by one
        process = True
        for line in file:
            # Process the line (for example, print it)
            ln = line.strip().split()
            if len(ln) > 0:
                if ln[0] == "verbatim" or ln[0] == "header":
                    process = not process
                if (
                    ln[0][0] != "%"
                    and ln[0][0] != "*"
                    and ln[0] != "header"
                    and ln[0] != "verbatim"
                    and process
                ):
                    data.append(ln)

        start_indices = []
        for i_d, d in enumerate(data):
            if len(d) == 3:
                if d[0] != "check" and d[0] != "%" and d[0] not in ignore_solutions:
                    if int(d[1]) > 1.1:
                        start_indices.append(i_d)

    n_data = len(data)

    noods = []
    for ind in start_indices:
        i = ind
        name = data[i][0]
        n_mbrs = int(data[i][1])
        logging.debug(f"Solution: {name}")
        logging.debug(f"Number of endmembers: {n_mbrs}")

        vars = data[i + 1 : i + n_mbrs]
        logging.debug(vars)
        i += n_mbrs

        mbr_proportions = []
        mbr_names = []
        for j in range(n_mbrs):
            n_lines = int(data[i][1])
            mbr_names.append(data[i][0].split("(")[1][:-1])
            mbr_proportions.append(data[i : i + n_lines])
            i += n_lines
        logging.debug(mbr_proportions)

        formulation = data[i][0]
        logging.debug(f"formulation: {formulation}")
        i += 1
        n_ints = int((n_mbrs * (n_mbrs - 1.0)) / 2.0)
        interactions = data[i : i + n_ints]
        logging.debug(interactions)

        nonideal_entropies = False
        nonideal_volumes = False

        We = [[0.0 for j in mbr_names[:-i]] for i in range(1, n_mbrs)]
        Ws = [[0.0 for j in mbr_names[:-i]] for i in range(1, n_mbrs)]
        Wv = [[0.0 for j in mbr_names[:-i]] for i in range(1, n_mbrs)]
        for ints in interactions:
            m = ints[0].split("(")[1].replace(")", "").split(",")
            m0 = mbr_names.index(m[0])
            m1 = mbr_names.index(m[1])
            We[m0][m1 - m0 - 1] = float(ints[1]) * 1.0e3
            Ws[m0][m1 - m0 - 1] = -float(ints[2])
            Wv[m0][m1 - m0 - 1] = float(ints[3]) * 1.0e-5

            if np.abs(Ws[m0][m1 - m0 - 1]) > 1.0e-10:
                nonideal_entropies = True
            if np.abs(Wv[m0][m1 - m0 - 1]) > 1.0e-10:
                nonideal_volumes = True

        i += n_ints

        if formulation == "asf":
            text_alphas = data[i : i + n_mbrs]
            logging.debug(text_alphas)
            i += n_mbrs

            Ae = [0.0 for i in range(n_mbrs)]
            As = [0.0 for i in range(n_mbrs)]
            Av = [0.0 for i in range(n_mbrs)]
            for alpha in text_alphas:
                try:
                    m = alpha[0].split("(")[1].replace(")", "").split(",")[0]
                except IndexError:
                    m = alpha[0]
                m0 = mbr_names.index(m)
                Ae[m0] = float(alpha[1])
                As[m0] = -float(alpha[2])
                Av[m0] = float(alpha[3]) * 1.0e6

        n_site_species = int(data[i][0])

        logging.debug(f"Number of site species: {n_site_species}")
        i += 1

        site_fractions = []
        for j in range(n_site_species):
            n_lines = int(data[i][1])
            site_fractions.append(data[i : i + n_lines])
            i += n_lines

        summed = 0.0
        sites = [[]]
        for site_species_and_expression in site_fractions:
            site_species = site_species_and_expression[0][0]
            expression = site_species_and_expression
            expression[0] = expression[0][2:]
            expression = reverse_polish(expression)
            summed += expression
            summed = sympify(summed)
            sites[-1].append(site_species)
            if summed == 1:
                summed = 0
                sites.append([])
        sites = sites[:-1]
        site_species = {}
        for n_site, site in enumerate(sites):
            for species in site:
                site_species[species] = n_site
        n_sites = len(sites)

        endmember_site_fractions_and_checks = []
        for j in range(n_mbrs):
            n_lines = 1
            while True:
                if (
                    data[i + n_lines][0] == "make"
                    or data[i + n_lines][0] == "delG(tran)"
                    or data[i + n_lines][0] == "check"
                    or data[i + n_lines][0] == "delG(od)"
                    or data[i + n_lines][0] == "delG(make)"
                    or data[i + n_lines][0] == "delG(mod)"
                    or data[i + n_lines][0] == "delG(rcal)"
                    or data[i + n_lines][0] == "DQF"
                ):
                    n_lines += 1
                else:
                    break
            endmember_site_fractions_and_checks.append(data[i : i + n_lines])
            i += n_lines

        mbr_initializations = []
        for f in endmember_site_fractions_and_checks:
            mbr = f[0][0]
            if len(f) == 1:
                mbr = f"{mbr_dataset}.{mbr}()"
            elif len(f) == 2:
                mbr = f"{mbr_dataset}.{mbr}()"
                assert f[1][0] == "check"
            else:
                make_idx = [g[0] for g in f].index("make")
                make = f[make_idx][1:]
                delG = [np.array(g[1:4], dtype=float) for g in f if "delG" in g[0]]
                delG = np.array([0.0, 0.0, 0.0]) if len(delG) == 0 else delG[0]
                delG *= [1.0e3, -1.0e3, 1.0e-5]
                n_make = int(make[0])
                el = 1
                combined_mbrs = "["
                combined_amounts = "["
                for j in range(n_make):
                    if make[el] == "ordered":
                        el += 1
                        m_string = make[el]
                        m_amount = float(parse_expr(make[el + 1]))
                        mineral = getattr(dataset, m_string)()
                        property_modifier = mineral.property_modifiers[0]
                        delG += ordering_modifier(property_modifier, True) * m_amount
                        combined_mbrs += f"{make[el]}_nood, "
                        noods.append(make[el])
                    elif make[el] == "disordered":
                        el += 1
                        m_string = make[el]
                        m_amount = float(parse_expr(make[el + 1]))
                        mineral = getattr(dataset, m_string)()
                        property_modifier = mineral.property_modifiers[0]
                        delG += ordering_modifier(property_modifier, False) * m_amount
                        combined_mbrs += f"{make[el]}_nood, "
                        noods.append(make[el])
                    elif make[el] == "equilibrium":
                        el += 1
                        combined_mbrs += f"{mbr_dataset}.{make[el]}(), "
                    else:
                        combined_mbrs += f"{mbr_dataset}.{make[el]}(), "
                    combined_amounts += f"{float(parse_expr(make[el+1]))}, "
                    el += 2
                combined_mbrs = combined_mbrs[:-2]
                combined_amounts = combined_amounts[:-2]
                combined_mbrs += "]"
                combined_amounts += "]"
                out_make += f'{mbr} = CombinedMineral({combined_mbrs}, {combined_amounts}, {list(delG)}, "{mbr}")\n'
            n_occs = int(f[0][2])
            site_occ = [[] for ln in range(n_sites)]
            site_occ_list = f[0][3:]
            ss = [[] for idx in range(n_sites)]
            site_amounts = [[] for idx in range(n_sites)]
            for j in range(n_occs):
                site_sp = site_occ_list[2 * j]
                i_site = site_species[site_sp]
                # rename for insertion into string
                if site_sp[0] == "x":
                    site_sp = site_sp[1:]
                    site_sp = "".join(filter(str.isalpha, site_sp)).title()

                ss[i_site].append(site_sp)
                site_amounts[i_site].append(parse_expr(site_occ_list[2 * j + 1]))
            total_site_amounts = [sum(amounts) for amounts in site_amounts]
            site_fractions = [
                [amount / total_site_amounts[j] for amount in amounts]
                for j, amounts in enumerate(site_amounts)
            ]

            site_formula = ""
            for m, site in enumerate(ss):
                site_formula += "["
                for n, species in enumerate(site):
                    site_formula += species
                    if site_fractions[m][n] != 1:
                        site_formula += str(site_fractions[m][n])
                site_formula += "]"
                if str(total_site_amounts[m]) != "1":
                    site_formula += str(total_site_amounts[m])

            mbr_initializations.append(f'[{mbr}, "{site_formula}"]')

        out_ss += f"class {name}(Solution):\n"
        out_ss += "    def __init__(self, molar_fractions=None):\n"
        out_ss += f'        self.name = "{name}"\n'
        if formulation == "asf":
            out_ss += "        self.solution_model = AsymmetricRegularSolution(\n"
        else:
            out_ss += "        self.solution_model = SymmetricRegularSolution(\n"
        out_ss += "            endmembers=[\n"
        for mbr_initialization in mbr_initializations:
            out_ss += f"                {mbr_initialization},\n"
        out_ss += "            ],\n"
        if formulation == "asf":
            out_ss += f"            alphas={Ae},\n"
        out_ss += f"            energy_interaction={We},\n"
        if nonideal_entropies:
            out_ss += f"            entropy_interaction={Ws},\n"
        if nonideal_volumes:
            out_ss += f"            volume_interaction={Wv},\n"

        out_ss += "        )\n"
        out_ss += "        Solution.__init__(self, molar_fractions=molar_fractions)\n\n"

        logging.debug(site_fractions)
        logging.debug(endmember_site_fractions_and_checks)
        logging.debug("")

    out_noods = ""
    noods = np.unique(sorted(noods))
    for nood in noods:
        mineral = getattr(dataset, nood)()
        params = mineral.params
        out_noods += f"{nood}_nood = Mineral({params})\n\n"

    # Preamble
    solution_dataset = solution_file.split("-")[-1].split(".")[0]
    underline = "^" * len(solution_dataset)
    preamble = (
        "# This file is part of BurnMan - a thermoelastic\n"
        "# and thermodynamic toolkit for the Earth and Planetary Sciences\n"
        f"# Copyright (C) 2012 - {current_year} by the BurnMan team, released under the GNU\n"
        "# GPL v2 or later.\n"
        "\n"
        '"""\n'
        f"{solution_dataset}\n"
        f"{underline}\n"
        "\n"
        "HPx-eos solutions using endmembers from\n"
        f"dataset {mbr_dataset}.\n"
        "The values in this document are all in S.I. units,\n"
        "unlike those in the original THERMOCALC file.\n"
        "This file is autogenerated using process_HPX_eos.py\n"
        '"""\n\n'
        f"from numpy import array, nan\n"
        f"from . import {mbr_dataset}\n"
        "from ..classes.mineral import Mineral\n"
        "from ..classes.solution import Solution\n"
        "from ..classes.solutionmodel import SymmetricRegularSolution\n"
        "from ..classes.solutionmodel import AsymmetricRegularSolution\n"
        "from ..classes.combinedmineral import CombinedMineral\n\n"
    )

    output_string = f"{preamble}\n{out_noods}\n{out_make}\n{out_ss}"

    output_file = f"../../minerals/{solution_dataset}.py"

    with open(output_file, "w") as text_file:
        print(f"from . import {solution_dataset}")
        text_file.write(output_string)

print("Copy the above to ../../minerals/__init__.py:")



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2024 by the BurnMan team, released under the GNU
# GPL v2 or later.

from burnman.eos.vinet import Vinet
from burnman.eos.mie_grueneisen_debye import MGDBase
from burnman.classes.calibrant import Calibrant
import numpy as np


"""
Dewaele_2012
^^^^^^^^^^^^
"""


class KBr_B2(Calibrant):
    """
    The B2 KBr pressure standard reported by
    Dewaele et al. (2012; https://doi.org/10.1103/PhysRevB.85.214105).
    """

    def __init__(self):
        def _pressure_Dewaele_KBr(volume, temperature, params):

            # Isothermal pressure (GPa)
            pressure_model = Vinet()
            P0 = pressure_model.pressure(params["T_0"], volume, params)

            # Thermal pressure
            Pth = (
                params["aK(V,T)"] + params["dK_dT_V"] * np.log(params["V_0"] / volume)
            ) * (temperature - params["T_0"])

            # Total pressure
            P = P0 + Pth

            return P

        _params_Dewaele_KBr = {
            "V_0": 3.8180e-5,
            "K_0": 14.9e9,
            "Kprime_0": 5.81,
            "aK(V,T)": 0.00222e9,
            "dK_dT_V": 0.0,
            "n": 2.0,
            "T_0": 300.0,
            "P_0": 0.0,
            "Z": 1.0,
        }

        Calibrant.__init__(self, _pressure_Dewaele_KBr, "pressure", _params_Dewaele_KBr)


class KCl_B2(Calibrant):
    """
    The B2 KCl pressure standard reported by
    Dewaele et al. (2012; https://doi.org/10.1103/PhysRevB.85.214105).
    """

    def __init__(self):
        def _pressure_Dewaele_KCl(volume, temperature, params):

            # Isothermal pressure (GPa)
            pressure_model = Vinet()
            P0 = pressure_model.pressure(params["T_0"], volume, params)

            # Thermal pressure
            Pth = (
                params["aK(V,T)"] + params["dK_dT_V"] * np.log(params["V_0"] / volume)
            ) * (temperature - params["T_0"])

            # Total pressure
            P = P0 + Pth

            return P

        _params_Dewaele_KCl = {
            "V_0": 3.28206e-5,
            "K_0": 17.2e9,
            "Kprime_0": 5.89,
            "aK(V,T)": 0.00224e9,
            "dK_dT_V": 0.0,
            "n": 2.0,
            "T_0": 300.0,
            "P_0": 0.0,
            "Z": 1.0,
        }

        Calibrant.__init__(self, _pressure_Dewaele_KCl, "pressure", _params_Dewaele_KCl)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2024 by the BurnMan team, released under the GNU
# GPL v2 or later.

from burnman.eos.birch_murnaghan import BirchMurnaghanBase as BM3
from burnman.classes.calibrant import Calibrant
import numpy as np

"""
Zhang_1999
^^^^^^^^^^
"""


class Fe_bcc(Calibrant):
    """
    The BCC Fe pressure standard reported by
    Zhang et al. (1999; https://dx.doi.org/10.1007/s002690050178).
    """

    def __init__(self):
        def _pressure_Zhang_bccFe(volume, temperature, params):

            # Isothermal pressure (GPa)
            pressure_model = BM3()
            P0 = pressure_model.pressure(params["T_0"], volume, params)

            # Thermal pressure
            Pth = (
                params["aK(V,T)"] + params["dK_dT_V"] * np.log(params["V_0"] / volume)
            ) * (temperature - params["T_0"])

            # Total pressure
            P = P0 + Pth

            return P

        _params_Zhang_bccFe = {
            "V_0": 7.08384e-06,
            "K_0": 155.0e9,
            "Kprime_0": 5.3,
            "aK(V,T)": 0.00648e9,
            "dK_dT_V": -0.022e9,
            "n": 1.0,
            "T_0": 300.0,
            "P_0": 0.0,
            "Z": 2.0,
        }

        Calibrant.__init__(self, _pressure_Zhang_bccFe, "pressure", _params_Zhang_bccFe)



from __future__ import absolute_import

# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

import numpy as np
import warnings
import pkgutil
from scipy.optimize import brentq
from scipy.special import binom

from . import equation_of_state as eos
from ..constants import gas_constant
from ..utils.math import bracket


class BroshCalphad(eos.EquationOfState):
    """
    Class for the high pressure CALPHAD equation of state by
    :cite:`Brosh2007`.
    """

    def volume(self, pressure, temperature, params):
        """
        Returns volume :math:`[m^3]` as a function of pressure :math:`[Pa]`.
        """
        X = [
            1.0
            / (
                1.0
                - params["a"][i - 2]
                + params["a"][i - 2]
                * np.power(
                    1.0 + i / (3.0 * params["a"][i - 2]) * pressure / params["K_0"],
                    1.0 / float(i),
                )
            )
            for i in range(2, 6)
        ]
        V_c = params["V_0"] * np.sum(
            [params["c"][i - 2] * np.power(X[i - 2], 3.0) for i in range(2, 6)]
        )

        nu = self._theta(pressure, params) / temperature
        dP = 1000.0
        dthetadP = (
            self._theta(pressure + dP / 2.0, params)
            - self._theta(pressure - dP / 2.0, params)
        ) / dP
        V_qh = (
            3.0
            * params["n"]
            * gas_constant
            * np.exp(-nu)
            / (1.0 - np.exp(-nu))
            * dthetadP
        )  # eq. 6

        f = np.sqrt(
            1.0
            + 2.0
            * params["b"][1]
            * (1.0 + params["delta"][1])
            * pressure
            / params["K_0"]
        )
        dIdP = (
            (1.0 + params["delta"][1])
            / (params["K_0"] * (1.0 + params["b"][1]))
            * np.exp((1.0 - f) / params["b"][1])
        )
        V_th = self._C_T(temperature, params) * dIdP

        # V = dG_c/dP + dG_qh/dP - C_T*(dI_P/dP)
        return V_c + V_qh + V_th

    def pressure(self, temperature, volume, params):
        def _delta_volume(pressure):
            return self.volume(pressure, temperature, params) - volume

        try:
            sol = bracket(_delta_volume, 300.0e9, 1.0e5, ())
        except ValueError:
            raise Exception(
                "Cannot find a pressure, perhaps you are outside "
                "the range of validity for the equation of state?"
            )

        return brentq(_delta_volume, sol[0], sol[1])

    def isothermal_bulk_modulus_reuss(self, pressure, temperature, volume, params):
        """
        Returns the isothermal bulk modulus :math:`K_T` :math:`[Pa]`
        as a function of pressure :math:`[Pa]`,
        temperature :math:`[K]` and volume :math:`[m^3]`.
        """
        dP = 1000.0
        dV = self.volume(pressure + dP / 2.0, temperature, params) - self.volume(
            pressure - dP / 2.0, temperature, params
        )

        return -volume * dP / dV

    def isentropic_bulk_modulus_reuss(self, pressure, temperature, volume, params):
        """
        Returns the adiabatic bulk modulus of the mineral. :math:`[Pa]`.
        """
        if temperature < 1.0e-10:
            return self.isothermal_bulk_modulus_reuss(
                pressure, temperature, volume, params
            )
        else:
            return (
                self.isothermal_bulk_modulus_reuss(
                    pressure, temperature, volume, params
                )
                * self.molar_heat_capacity_p(pressure, temperature, volume, params)
                / self.molar_heat_capacity_v(pressure, temperature, volume, params)
            )

    def shear_modulus(self, pressure, temperature, volume, params):
        """
        Returns the shear modulus :math:`G` of the mineral. :math:`[Pa]`
        """
        return 0.0

    def molar_internal_energy(self, pressure, temperature, volume, params):
        """
        Returns the internal energy of the mineral. :math:`[J/mol]`
        """

        return (
            self.gibbs_free_energy(pressure, temperature, volume, params)
            - pressure * self.volume(pressure, temperature, params)
            + temperature * self.entropy(pressure, temperature, volume, params)
        )

    def _Cp_1bar(self, temperature, params):
        # first, identify which of the piecewise segments we're in
        i = np.argmax(
            [T > temperature for T in list(zip(*params["gibbs_coefficients"]))[0]]
        )

        # select the appropriate coefficients
        coeffs = params["gibbs_coefficients"][i][1]
        Cp = -(
            coeffs[2]
            + 2.0 * coeffs[3] / temperature / temperature
            + 6.0 * coeffs[4] / (temperature * temperature * temperature)
            + 12.0 * coeffs[5] * np.power(temperature, -4.0)
            + 90.0 * coeffs[6] * np.power(temperature, -10.0)
            + 2.0 * coeffs[7] * temperature
            + 6.0 * coeffs[8] * temperature * temperature
            + 12.0 * coeffs[9] * temperature * temperature * temperature
            + 42.0 * coeffs[10] * np.power(temperature, 6.0)
            - 0.25 * coeffs[11] / np.sqrt(temperature)
            - coeffs[12] / temperature
        )
        return Cp

    def _S_1bar(self, temperature, params):
        # first, identify which of the piecewise segments we're in
        i = np.argmax(
            [T > temperature for T in list(zip(*params["gibbs_coefficients"]))[0]]
        )

        # select the appropriate coefficients
        coeffs = params["gibbs_coefficients"][i][1]
        S = -(
            coeffs[1]
            + coeffs[2] * (1.0 + np.log(temperature))
            - coeffs[3] / temperature / temperature
            - 2.0 * coeffs[4] / (temperature * temperature * temperature)
            - 3.0 * coeffs[5] * np.power(temperature, -4.0)
            - 9.0 * coeffs[6] * np.power(temperature, -10.0)
            + 2.0 * coeffs[7] * temperature
            + 3.0 * coeffs[8] * temperature * temperature
            + 4.0 * coeffs[9] * temperature * temperature * temperature
            + 7.0 * coeffs[10] * np.power(temperature, 6.0)
            + 0.5 * coeffs[11] / np.sqrt(temperature)
            + coeffs[12] / temperature
        )
        return S

    def _gibbs_1bar(self, temperature, params):
        # first, identify which of the piecewise segments we're in
        i = np.argmax(
            [T > temperature for T in list(zip(*params["gibbs_coefficients"]))[0]]
        )

        # select the appropriate coefficients
        coeffs = params["gibbs_coefficients"][i][1]
        gibbs = (
            coeffs[0]
            + coeffs[1] * temperature
            + coeffs[2] * temperature * np.log(temperature)
            + coeffs[3] / temperature
            + coeffs[4] / (temperature * temperature)
            + coeffs[5] / (temperature * temperature * temperature)
            + coeffs[6] * np.power(temperature, -9.0)
            + coeffs[7] * temperature * temperature
            + coeffs[8] * temperature * temperature * temperature
            + coeffs[9] * np.power(temperature, 4.0)
            + coeffs[10] * np.power(temperature, 7.0)
            + coeffs[11] * np.sqrt(temperature)
            + coeffs[12] * np.log(temperature)
        )
        return gibbs

    def _X(self, pressure, params):
        return [
            1.0
            / (
                1.0
                - params["a"][n - 2]
                + params["a"][n - 2]
                * np.power(
                    1.0
                    + float(n) / (3.0 * params["a"][n - 2]) * pressure / params["K_0"],
                    1.0 / float(n),
                )
            )
            for n in range(2, 6)
        ]  # eq. A2

    def _Gamma(self, n, an, Xn):
        def d(k, Xn):
            return (
                np.power(Xn, 3.0 - float(k)) * float(k) / (float(k) - 3.0)
                if k != 3
                else -3.0 * np.log(Xn)
            )  # eq. A9

        return (
            3.0
            * np.power(an, 1.0 - float(n))
            / float(n)
            * np.sum(
                [
                    binom(n, k) * np.power(an - 1.0, float(n - k)) * d(k, Xn)
                    for k in range(0, n + 1)
                ]
            )
        )  # eq. A9, CHECKED

    def _theta(self, pressure, params):
        # Theta (for quasiharmonic term)
        ab2 = 1.0 / (3.0 * params["b"][0] - 1.0)
        K0b = params["K_0"] / (1.0 + params["delta"][0])  # eq. B1b
        XT2 = 1.0 / (
            1.0 - ab2 + ab2 * np.power(1.0 + 2.0 / (3.0 * ab2) * pressure / K0b, 0.5)
        )  # eq. 6 b of SE2015

        # eq. B1 (6 of SE2015)
        return params["theta_0"] * np.exp(
            params["grueneisen_0"]
            / (1.0 + params["delta"][0])
            * (self._Gamma(2, ab2, XT2) - self._Gamma(2, ab2, 1.0))
        )

    def _interpolating_function(self, pressure, params):
        f = np.sqrt(
            1.0
            + 2.0
            * params["b"][1]
            * (1.0 + params["delta"][1])
            * pressure
            / params["K_0"]
        )

        # eq. D2 (9 of SE2015)
        return (
            1.0
            / (1.0 + params["b"][1])
            * (params["b"][1] + f)
            * np.exp((1.0 - f) / params["b"][1])
        )

    def _gibbs_qh(self, temperature, theta, n):
        return (
            3.0
            * n
            * gas_constant
            * temperature
            * np.log(1.0 - np.exp(-theta / temperature))
        )  # eq. 5

    def _S_qh(self, temperature, theta, n):
        nu = theta / temperature
        return (
            3.0
            * n
            * gas_constant
            * (nu / (np.exp(nu) - 1.0) - np.log(1.0 - np.exp(-nu)))
        )

    def _C_T(self, temperature, params):
        # C, which is the (G_qh(t,p0) - G_sgte(t,p0)) term
        G_SGTE = self._gibbs_1bar(temperature, params)
        G_qh0 = self._gibbs_qh(temperature, params["theta_0"], params["n"])
        if temperature < params["T_0"]:
            C_T = (
                temperature * temperature / (2.0 * params["T_0"]) * params["delta_Cpr"]
            )

        else:
            C_T = (
                (G_qh0 - G_SGTE)
                + params["delta_Gr"]
                - (temperature - params["T_0"]) * params["delta_Sr"]
                + (temperature - params["T_0"] / 2.0) * params["delta_Cpr"]
            )

        return C_T

    def gibbs_free_energy(self, pressure, temperature, volume, params):
        """
        Returns the Gibbs free energy of the mineral. :math:`[J/mol]`
        """
        # Cold compression term, eq. A8
        X = self._X(pressure, params)
        G_c = (
            params["K_0"]
            * params["V_0"]
            * np.sum(
                [
                    params["c"][n - 2]
                    * (
                        self._Gamma(n, params["a"][n - 2], X[n - 2])
                        - self._Gamma(n, params["a"][n - 2], 1.0)
                    )
                    for n in range(2, 6)
                ]
            )
        )

        # G_SGTE
        G_SGTE = self._gibbs_1bar(temperature, params)

        # G_qh
        theta = self._theta(pressure, params)
        G_qh = self._gibbs_qh(temperature, theta, params["n"])
        G_qh0 = self._gibbs_qh(temperature, params["theta_0"], params["n"])

        C_T = self._C_T(temperature, params)
        I_P = self._interpolating_function(pressure, params)
        return G_SGTE + G_c + G_qh - G_qh0 + C_T * (1.0 - I_P)

    def entropy(self, pressure, temperature, volume, params):
        """
        Returns the molar entropy of the mineral. :math:`[J/K/mol]`
        """

        S_SGTE = self._S_1bar(temperature, params)

        # S_qh
        theta = self._theta(pressure, params)
        S_qh = self._S_qh(temperature, theta, params["n"])
        S_qh0 = self._S_qh(temperature, params["theta_0"], params["n"])

        # dCdT, which is the (S_qh(t,p0) - S_sgte(t,p0)) term
        if temperature < params["T_0"]:
            dC_TdT = temperature / params["T_0"] * params["delta_Cpr"]

        else:
            dC_TdT = -(S_qh0 - S_SGTE) - params["delta_Sr"] + params["delta_Cpr"]

        I_P = self._interpolating_function(pressure, params)
        return S_SGTE + S_qh - S_qh0 - dC_TdT * (1.0 - I_P)

    def molar_heat_capacity_p(self, pressure, temperature, volume, params):
        """
        Returns the molar isobaric heat capacity :math:`[J/K/mol]`.
        For now, this is calculated by numerical differentiation.
        """
        dT = 0.1
        if temperature < dT / 2.0:
            return 0.0
        else:
            dS = self.entropy(
                pressure, temperature + dT / 2.0, volume, params
            ) - self.entropy(pressure, temperature - dT / 2.0, volume, params)
            return temperature * dS / dT

    def thermal_expansivity(self, pressure, temperature, volume, params):
        """
        Returns the volumetric thermal expansivity :math:`[1/K]`.
        For now, this is calculated by numerical differentiation.
        """
        dT = 0.1
        if temperature < dT / 2.0:
            return 0.0
        else:
            dV = self.volume(pressure, temperature + dT / 2.0, params) - self.volume(
                pressure, temperature - dT / 2.0, params
            )
            return dV / dT / volume

    def grueneisen_parameter(self, pressure, temperature, volume, params):
        """
        Returns the grueneisen parameter.
        This is a dependent thermodynamic variable in this equation of state.
        """
        Cv = self.molar_heat_capacity_v(pressure, temperature, volume, params)
        if Cv == 0.0:
            return 0.0
        else:
            return (
                self.thermal_expansivity(pressure, temperature, volume, params)
                * self.isothermal_bulk_modulus_reuss(
                    pressure, temperature, volume, params
                )
                * self.volume(pressure, temperature, params)
            ) / Cv

    def calculate_transformed_parameters(self, params):
        """
        This function calculates the "c" parameters of the :cite:`Brosh2007`
        equation of state.
        """
        Zs = pkgutil.get_data("burnman", "data/input_masses/atomic_numbers.dat")
        Zs = Zs.decode("ascii").split("\n")
        Z = {
            str(sl[0]): int(sl[1])
            for sl in [line.split() for line in Zs if len(line) > 0 and line[0] != "#"]
        }

        nZs = [(n_at, float(Z[el])) for (el, n_at) in params["formula"].items()]

        # eq. A2 at 300 TPa
        X3_300TPa = [
            np.power(
                1.0
                - params["a"][i - 2]
                + params["a"][i - 2]
                * np.power(
                    (
                        1.0
                        + float(i)
                        / (3.0 * params["a"][i - 2])
                        * 300.0e12
                        / params["K_0"]
                    ),
                    1.0 / float(i),
                ),
                -3.0,
            )
            for i in range(2, 6)
        ]

        # eq. A2 at 330 TPa
        X3_330TPa = [
            np.power(
                1.0
                - params["a"][i - 2]
                + params["a"][i - 2]
                * np.power(
                    (
                        1.0
                        + float(i)
                        / (3.0 * params["a"][i - 2])
                        * 330.0e12
                        / params["K_0"]
                    ),
                    1.0 / float(i),
                ),
                -3.0,
            )
            for i in range(2, 6)
        ]

        # eq. A6a, m^3/mol
        V_QSM_300TPa = (
            np.sum(
                [
                    n_at
                    * (
                        0.02713
                        * np.exp(
                            0.97626 * np.log(Zi) - 0.057848 * np.log(Zi) * np.log(Zi)
                        )
                    )
                    for (n_at, Zi) in nZs
                ]
            )
            * 1.0e-6
        )

        # eq. A6b, m^3/mol
        V_QSM_330TPa = (
            np.sum(
                [
                    n_at
                    * (
                        0.025692
                        * np.exp(
                            0.97914 * np.log(Zi) - 0.057741 * np.log(Zi) * np.log(Zi)
                        )
                    )
                    for (n_at, Zi) in nZs
                ]
            )
            * 1.0e-6
        )

        A = np.array(
            [
                [1.0, 1.0, 1.0, 1.0],  # eq A3
                [0.0, 6.0, 8.0, 9.0],  # eq A4
                X3_300TPa,  # eq A5a
                X3_330TPa,
            ]
        )  # eq A5b

        b = np.array(
            [1.0, 8.0, V_QSM_300TPa / params["V_0"], V_QSM_330TPa / params["V_0"]]
        )

        # does not quite reproduce the published values of c
        # A.c consistently gives b[2], b[3] ~1% larger than Brosh
        return np.linalg.solve(A, b)

    def validate_parameters(self, params):
        """
        Check for existence and validity of the parameters
        """

        params["T_0"] = 298.15
        if "P_0" not in params:
            params["P_0"] = 1.0e5

        if "a" not in params:
            params["a"] = [
                (float(i) - 1.0) / (3.0 * params["Kprime_0"] - 1.0) for i in range(2, 6)
            ]  # eq. A2

        if "c" not in params:
            params["c"] = self.calculate_transformed_parameters(params)

        # Calculate reference values for gibbs free energy and heat capacity
        nur = params["theta_0"] / params["T_0"]
        G_qhr = self._gibbs_qh(params["T_0"], params["theta_0"], params["n"])
        S_qhr = self._S_qh(params["T_0"], params["theta_0"], params["n"])
        Cp_qhr = (
            3.0
            * params["n"]
            * gas_constant
            * nur
            * nur
            * np.exp(nur)
            / np.power(np.exp(nur) - 1.0, 2.0)
        )

        G_SGTEr = self._gibbs_1bar(params["T_0"], params)
        S_SGTEr = self._S_1bar(params["T_0"], params)
        Cp_SGTEr = self._Cp_1bar(params["T_0"], params)

        params["delta_Cpr"] = Cp_SGTEr - Cp_qhr
        params["delta_Gr"] = G_SGTEr - G_qhr
        params["delta_Sr"] = S_SGTEr - S_qhr

        # Check that all the required keys are in the dictionary
        expected_keys = [
            "gibbs_coefficients",
            "V_0",
            "K_0",
            "Kprime_0",
            "theta_0",
            "grueneisen_0",
            "delta",
            "b",
        ]
        for k in expected_keys:
            if k not in params:
                raise KeyError("params object missing parameter : " + k)

        # Finally, check that the values are reasonable.
        if params["P_0"] < 0.0:
            warnings.warn("Unusual value for P_0", stacklevel=2)
        if params["V_0"] < 1.0e-7 or params["V_0"] > 1.0e-3:
            warnings.warn("Unusual value for V_0", stacklevel=2)
        if params["K_0"] < 1.0e9 or params["K_0"] > 1.0e13:
            warnings.warn("Unusual value for K_0", stacklevel=2)
        if params["Kprime_0"] < 0.0 or params["Kprime_0"] > 10.0:
            warnings.warn("Unusual value for Kprime_0", stacklevel=2)



from __future__ import print_function

# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2018 by the BurnMan team, released under the GNU
# GPL v2 or later.

from copy import deepcopy
import numpy as np
from scipy.optimize import nnls

from ..utils.chemistry import dictionarize_formula, formula_mass
from ..utils.misc import OrderedCounter


def file_to_composition_list(fname, unit_type, normalize):
    """
    Takes an input file with a specific format and returns a list of
    compositions (and associated comments) contained in that file.

    :param fname: Path to ascii file containing composition data.
        Lines beginning with a hash are not read.
        The first read-line of the datafile contains a list of tab or
        space-separated components (e.g. FeO or SiO2), followed by the
        word Comment.
        Following lines are lists of floats with the amounts of each component.
        After the component amounts, the user can write anything they like
        in the Comment section.
    :type fname: str
    :param unit_type: 'mass', 'weight' or 'molar'
        Specify whether the compositions in the file are given as
        mass (weight) or molar amounts.
    :type unit_type: str
    :param normalize: If False, absolute numbers of moles/grams of component are stored,
        otherwise the component amounts of returned compositions will
        sum to one (until Composition.renormalize() is used).
    :type normalize : bool

    """
    lines = list(
        filter(
            None, [line.rstrip("\n").split() for line in open(fname) if line[0] != "#"]
        )
    )
    n_components = lines[0].index("Comment")
    components = lines[0][:n_components]
    comments = [line[n_components:] for line in lines[1:]]
    compositions = np.array(
        [map(float, ln) for ln in list(zip(*(list(zip(*lines[1:]))[:n_components])))]
    )
    return [
        Composition(OrderedCounter(dict(zip(components, c))), unit_type, normalize)
        for c in compositions
    ], comments


class Composition(object):
    """
    Class for a composition object, which can be used
    to store, modify and renormalize compositions,
    and also convert between mass, molar
    and atomic amounts. Weight is provided as an alias
    for mass, as we assume that only Earthlings
    will use this software.

    This class is available as ``burnman.Composition``.
    """

    def __init__(self, composition_dictionary, unit_type="mass", normalize=False):
        """
        Create a composition using a dictionary and unit type.

        :param composition_dictionary: Dictionary of components
            (given as a string) and their amounts.
        :type composition_dictionary: dictionary
        :param unit_type: 'mass', 'weight' or 'molar' (optional, 'mass' as default)
            Specify whether the input composition is given as mass or
            molar amounts.
        :type unit_type: str
        :param normalize: If False, absolute numbers of kilograms/moles of component are
            stored, otherwise the component amounts of returned compositions
            will sum to one (until Composition.renormalize() is used).
        :type normalize: bool
        """

        self._cached = {}

        n_total = float(sum(composition_dictionary.values()))

        # Create the input dictionary, normalize if requested
        input_dictionary = OrderedCounter(deepcopy(composition_dictionary))
        if normalize:
            for k in composition_dictionary.keys():
                input_dictionary[k] = composition_dictionary[k] / n_total

        # Break component formulae into atomic dictionaries
        self.component_formulae = {
            c: dictionarize_formula(c) for c in composition_dictionary.keys()
        }

        # Create lists of elemental compositions of components
        self.element_list = OrderedCounter()
        for component in self.component_formulae.values():
            self.element_list += OrderedCounter(
                {element: n_atoms for (element, n_atoms) in component.items()}
            )
        self.element_list = list(self.element_list.keys())

        if unit_type == "mass" or unit_type == "weight":
            self.mass_composition = input_dictionary
        elif unit_type == "molar":
            self.mass_composition = self._mole_to_mass_composition(input_dictionary)
        else:
            raise Exception(
                "Unit type not yet implemented. "
                "Should be either mass, weight or molar."
            )

    def renormalize(self, unit_type, normalization_component, normalization_amount):
        """
        Change the total amount of material in the composition
        to satisfy a given normalization condition
        (mass, weight, molar, or atomic)

        :param unit_type: 'mass', 'weight', 'molar' or 'atomic'
            Unit type with which to normalize the composition
        :type unit_type: str
        :param normalization_component: Component/element on which to renormalize.
            String must either be one of the components/elements
            already in the composition, or have the value 'total'.
        :type normalization_component: str
        :param normalization_amount: Amount of component in the
            renormalised composition.
        :type normalization_amount: float
        """

        if unit_type not in ["mass", "weight", "molar", "atomic"]:
            raise Exception(
                "unit_type not yet implemented."
                "Should be either mass, weight, molar or atomic."
            )

        c = self.composition(unit_type)

        if normalization_component == "total":
            f = normalization_amount / float(sum(c.values()))
        else:
            f = normalization_amount / c[normalization_component]

        new_mass_composition = OrderedCounter()
        for k in self.mass_composition.keys():
            new_mass_composition[k] = self.mass_composition[k] * f

        self.mass_composition = new_mass_composition

    def add_components(self, composition_dictionary, unit_type):
        """
        Add (or remove) components from the composition.
        The components are added to the current state of the
        (mass, weight or molar) composition; if the composition has
        been renormalised, then this should be taken into account.

        :param composition_dictionary: Components to add, and their amounts.
        :type composition_dictionary: dictionary
        :param unit_type: 'mass', 'weight' or 'molar'.
            Unit type of the components to be added.
        :type unit_type: str
        """
        if unit_type == "mass" or unit_type == "weight":
            composition = self.mass_composition
        elif unit_type == "molar":
            composition = self.molar_composition
        else:
            raise Exception(
                "Unit type not recognised. " "Should be either mass, weight or molar."
            )

        composition += OrderedCounter(composition_dictionary)

        # Reinitialize composition object
        self.__init__(composition, unit_type)

    def change_component_set(self, new_component_list):
        """
        Change the set of basis components without
        changing the bulk composition.

        Will raise an exception if the new component set is
        invalid for the given composition.

        :param new_component_list: New set of basis components.
        :type new_component_list: list of strings
        """
        composition = np.array(
            [self.atomic_composition[element] for element in self.element_list]
        )
        component_matrix = np.zeros((len(new_component_list), len(self.element_list)))

        for i, component in enumerate(new_component_list):
            formula = dictionarize_formula(component)
            for element, n_atoms in formula.items():
                component_matrix[i][self.element_list.index(element)] = n_atoms

        sol = nnls(component_matrix.T, composition)
        if sol[1] < 1.0e-10:
            component_amounts = sol[0]
        else:
            raise Exception(
                "Failed to change component set. "
                "Could not find a non-negative "
                f"least squares solution (residual={sol[1]}). "
                "Can the bulk composition be described "
                "with this set of components?"
            )

        composition = OrderedCounter(dict(zip(new_component_list, component_amounts)))

        # Reinitialize the object
        self.__init__(composition, "molar")

    def _mole_to_mass_composition(self, molar_comp):
        """
        Hidden function to returns the mass composition as a counter [kg]
        """
        cf = self.component_formulae
        mass_composition = OrderedCounter(
            {c: molar_comp[c] * formula_mass(cf[c]) for c in molar_comp.keys()}
        )

        return mass_composition

    @property
    def weight_composition(self):
        """
        An alias for mass composition [kg].
        """
        return self.mass_composition

    @property
    def molar_composition(self):
        """
        Returns the molar composition as a counter [moles]
        """
        mass_comp = self.mass_composition
        cf = self.component_formulae

        return OrderedCounter(
            {c: mass_comp[c] / formula_mass(cf[c]) for c in mass_comp.keys()}
        )

    @property
    def atomic_composition(self):
        """
        Returns the atomic composition as a counter [moles]
        """

        return self._moles_to_atoms(self.molar_composition)

    def composition(self, unit_type):
        """
        Helper function to return the composition in the
        desired type.

        :param unit_type: One of 'mass', 'weight', 'molar' and 'atomic'.
        :type unit_type: str

        :returns: Mass (weight), molar or atomic composition.
        :rtype: OrderedCounter
        """
        return getattr(self, f"{unit_type}_composition")

    def _moles_to_atoms(self, molar_comp_dictionary):
        """
        Hidden function that converts a molar component
        dictionary into an atomic (elemental) dictionary
        """
        component_matrix = np.zeros(
            (len(self.component_formulae), len(self.element_list))
        )
        cf = self.component_formulae
        molar_composition_vector = np.zeros(len(cf))
        for i, (component, formula) in enumerate(cf.items()):
            molar_composition_vector[i] = molar_comp_dictionary[component]

            for element, n_atoms in formula.items():
                component_matrix[i][self.element_list.index(element)] = n_atoms

        atom_compositions = np.dot(molar_composition_vector, component_matrix)
        return OrderedCounter(dict(zip(self.element_list, atom_compositions)))

    def print(
        self,
        unit_type,
        significant_figures=1,
        normalization_component="total",
        normalization_amount=None,
    ):
        """
        Pretty-print function for the composition
        This does not renormalize the Composition object itself,
        only the printed values.

        :param unit_type: 'mass', 'weight', 'molar' or 'atomic'
            Unit type in which to print the composition.
        :type unit_type: str
        :param significant_figures: Number of significant figures
            for each amount.
        :type significant_figures: int
        :param normalization_component: Component/element on which to renormalize.
            String must either be one of the components/elements
            already in composite, or have the value 'total'.
            (default = 'total')
        :type normalization_component: str
        :param normalization_amount: Amount of component in the
            renormalised composition. If not explicitly set,
            no renormalization will be applied.
        :type normalization_amount: float
        """

        if unit_type not in ["mass", "weight", "molar", "atomic"]:
            raise Exception(
                "unit_type not yet implemented."
                "Should be either mass, weight, molar or atomic."
            )

        c = self.composition(unit_type)
        print(f"{unit_type.capitalize()} composition")

        if normalization_amount is None:
            f = 1
        elif normalization_component == "total":
            f = normalization_amount / float(sum(c.values()))
        else:
            f = normalization_amount / c[normalization_component]

        for key, value in sorted(c.items()):
            print(f"{key}: {value*f:0.{significant_figures}f}")



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2024 by the BurnMan team, released under the GNU
# GPL v2 or later.

"""
Holmes_1989
^^^^^^^^^^^
"""

from burnman.classes.calibrant import Calibrant
import numpy as np
from burnman.utils.unitcell import molar_volume_from_unit_cell_volume


class Pt(Calibrant):
    """
    The Pt pressure standard reported by
    Holmes et al. (1989; https://doi.org/10.1063/1.344177).
    """

    def __init__(self):
        def _pressure(volume, temperature, params):
            X = np.power(volume / params["V_0"], 1.0 / 3.0)
            P_300 = (
                3.0
                * params["beta_T"]
                * (1.0 - X)
                / (X * X)
                * np.exp(params["eta"] * (1.0 - X))
            )

            return P_300 + params["alpha_T"] * params["beta_T"] * (temperature - 300.0)

        Z = 4.0
        _params = {
            "V_0": molar_volume_from_unit_cell_volume(60.38, Z),
            "beta_T": 798.31e9 / 3.0,
            "eta": 7.2119,
            "beta_prime_T": (7.2119 / 1.5) + 1.0,
            "alpha_T": 2.61e-5,
            "Z": Z,
        }

        Calibrant.__init__(self, _pressure, "pressure", _params)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit
# for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2024 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import

import numpy as np
import scipy.optimize as opt
import warnings

# Try to import the jit from numba.  If it is
# not available, just go with the standard
# python interpreter
try:
    from numba import jit
except ImportError:

    def jit(fn):
        return fn


from . import birch_murnaghan as bm
from . import debye
from . import equation_of_state as eos
from . import bukowinski_electronic as el
from ..utils.math import bracket


@jit(nopython=True)
def _grueneisen_parameter_slb(V_0, volume, gruen_0, q_0):
    """global function with plain parameters so jit will work"""
    x = V_0 / volume
    f = 1.0 / 2.0 * (pow(x, 2.0 / 3.0) - 1.0)
    a1_ii = 6.0 * gruen_0  # EQ 47
    a2_iikk = -12.0 * gruen_0 + 36.0 * gruen_0 * gruen_0 - 18.0 * q_0 * gruen_0  # EQ 47
    nu_o_nu0_sq = 1.0 + a1_ii * f + (1.0 / 2.0) * a2_iikk * f * f  # EQ 41
    return 1.0 / 6.0 / nu_o_nu0_sq * (2.0 * f + 1.0) * (a1_ii + a2_iikk * f)


@jit(nopython=True)
def _delta_pressure(
    x,
    pressure,
    temperature,
    V_0,
    T_0,
    Debye_0,
    n,
    a1_ii,
    a2_iikk,
    b_iikk,
    b_iikkmm,
    bel_0,
    gel,
):
    f = 0.5 * (pow(V_0 / x, 2.0 / 3.0) - 1.0)
    nu_o_nu0_sq = 1.0 + a1_ii * f + 1.0 / 2.0 * a2_iikk * f * f
    debye_temperature = Debye_0 * np.sqrt(nu_o_nu0_sq)
    E_th = debye.thermal_energy(
        temperature, debye_temperature, n
    )  # thermal energy at temperature T
    E_th_ref = debye.thermal_energy(
        T_0, debye_temperature, n
    )  # thermal energy at reference temperature
    nu_o_nu0_sq = 1.0 + a1_ii * f + (1.0 / 2.0) * a2_iikk * f * f  # EQ 41
    gr = 1.0 / 6.0 / nu_o_nu0_sq * (2.0 * f + 1.0) * (a1_ii + a2_iikk * f)

    Pel = (
        0.5
        * gel
        * bel_0
        * np.power(x / V_0, gel)
        * (temperature * temperature - T_0 * T_0)
        / x
    )

    return (
        (1.0 / 3.0)
        * (pow(1.0 + 2.0 * f, 5.0 / 2.0))
        * ((b_iikk * f) + (0.5 * b_iikkmm * f * f))
        + gr * (E_th - E_th_ref) / x
        + Pel
        - pressure
    )  # EQ 21


class SLBBase(eos.EquationOfState):
    """
    Base class for the finite strain-Mie-Grueneiesen-Debye equation of state
    detailed in :cite:`Stixrude2005`.  For the most part the equations are
    all third order in strain, but see further the :class:`burnman.slb.SLB2`
    and :class:`burnman.slb.SLB3` classes.
    """

    def _debye_temperature(self, x, params):
        """
        Finite strain approximation for Debye Temperature [K]
        x = ref_vol/vol
        """
        f = 1.0 / 2.0 * (pow(x, 2.0 / 3.0) - 1.0)
        a1_ii = 6.0 * params["grueneisen_0"]  # EQ 47
        a2_iikk = (
            -12.0 * params["grueneisen_0"]
            + 36.0 * pow(params["grueneisen_0"], 2.0)
            - 18.0 * params["q_0"] * params["grueneisen_0"]
        )  # EQ 47
        nu_o_nu0_sq = 1.0 + a1_ii * f + 1.0 / 2.0 * a2_iikk * f * f
        if nu_o_nu0_sq > 0.0:
            return params["Debye_0"] * np.sqrt(nu_o_nu0_sq)
        else:
            raise Exception(
                f"This volume (V = {1./x:.2f}*V_0) exceeds the "
                "valid range of the thermal "
                "part of the slb equation of state."
            )

    def volume_dependent_q(self, x, params):
        """
        Finite strain approximation for :math:`q`, the isotropic volume strain
        derivative of the grueneisen parameter.
        """
        f = 1.0 / 2.0 * (pow(x, 2.0 / 3.0) - 1.0)
        a1_ii = 6.0 * params["grueneisen_0"]  # EQ 47
        a2_iikk = (
            -12.0 * params["grueneisen_0"]
            + 36.0 * pow(params["grueneisen_0"], 2.0)
            - 18.0 * params["q_0"] * params["grueneisen_0"]
        )  # EQ 47
        nu_o_nu0_sq = 1.0 + a1_ii * f + (1.0 / 2.0) * a2_iikk * f * f  # EQ 41
        gr = 1.0 / 6.0 / nu_o_nu0_sq * (2.0 * f + 1.0) * (a1_ii + a2_iikk * f)
        # avoids divide by zero if grueneisen_0 = 0.
        if np.abs(params["grueneisen_0"]) < 1.0e-10:
            q = 1.0 / 9.0 * (18.0 * gr - 6.0)
        else:
            q = (
                1.0
                / 9.0
                * (
                    18.0 * gr
                    - 6.0
                    - 1.0
                    / 2.0
                    / nu_o_nu0_sq
                    * (2.0 * f + 1.0)
                    * (2.0 * f + 1.0)
                    * a2_iikk
                    / gr
                )
            )
        return q

    def _isotropic_eta_s(self, x, params):
        """
        Finite strain approximation for :math:`eta_{s0}`, the isotropic shear
        strain derivative of the grueneisen parameter.
        """
        f = 1.0 / 2.0 * (pow(x, 2.0 / 3.0) - 1.0)
        a2_s = -2.0 * params["grueneisen_0"] - 2.0 * params["eta_s_0"]  # EQ 47
        a1_ii = 6.0 * params["grueneisen_0"]  # EQ 47
        a2_iikk = (
            -12.0 * params["grueneisen_0"]
            + 36.0 * pow(params["grueneisen_0"], 2.0)
            - 18.0 * params["q_0"] * params["grueneisen_0"]
        )  # EQ 47
        nu_o_nu0_sq = 1.0 + a1_ii * f + (1.0 / 2.0) * a2_iikk * pow(f, 2.0)  # EQ 41
        gr = 1.0 / 6.0 / nu_o_nu0_sq * (2.0 * f + 1.0) * (a1_ii + a2_iikk * f)
        # EQ 46 NOTE the typo from Stixrude 2005:
        eta_s = -gr - (
            1.0 / 2.0 * pow(nu_o_nu0_sq, -1.0) * pow((2.0 * f) + 1.0, 2.0) * a2_s
        )

        return eta_s

    def volume(self, pressure, temperature, params):
        """
        Returns molar volume. :math:`[m^3]`
        """
        T_0 = params["T_0"]
        Debye_0 = params["Debye_0"]
        V_0 = params["V_0"]
        dV = 1.0e-2 * params["V_0"]
        n = params["n"]

        a1_ii = 6.0 * params["grueneisen_0"]  # EQ 47
        a2_iikk = (
            -12.0 * params["grueneisen_0"]
            + 36.0 * pow(params["grueneisen_0"], 2.0)
            - 18.0 * params["q_0"] * params["grueneisen_0"]
        )  # EQ 47

        b_iikk = 9.0 * params["K_0"]  # EQ 28
        b_iikkmm = 27.0 * params["K_0"] * (params["Kprime_0"] - 4.0)  # EQ 29z

        bel_0, gel = 0.0, 1.0
        if self.conductive:
            bel_0, gel = params["bel_0"], params["gel"]

        # Finding the volume at a given pressure requires a
        # root-finding scheme. Here we use brentq to find the root.

        # Root-finding using brentq requires bounds to be specified.
        # We do this using a bracketing function.
        args = (
            pressure,
            temperature,
            V_0,
            T_0,
            Debye_0,
            n,
            a1_ii,
            a2_iikk,
            b_iikk,
            b_iikkmm,
            bel_0,
            gel,
        )

        try:
            # The first attempt to find a bracket for
            # root finding uses V_0 as a starting point
            sol = bracket(_delta_pressure, V_0, dV, args)
        except Exception:
            # At high temperature, the naive bracketing above may
            # try a volume guess that exceeds the point at which the
            # bulk modulus goes negative at that temperature.
            # In this case, we try a more nuanced approach by
            # first finding the volume at which the bulk modulus goes
            # negative, and then either (a) raising an exception if the
            # desired pressure is less than the pressure at that volume,
            # or (b) using that pressure to create a better bracket for
            # brentq.
            def _K_T(V, T, params):
                return self.isothermal_bulk_modulus_reuss(0.0, T, V, params)

            sol_K_T = bracket(_K_T, V_0, dV, args=(temperature, params))
            V_crit = opt.brentq(
                _K_T, sol_K_T[0], sol_K_T[1], args=(temperature, params)
            )
            P_min = self.pressure(temperature, V_crit, params)
            if P_min > pressure:
                raise Exception(
                    "The desired pressure is not achievable "
                    "at this temperature. The minimum pressure "
                    f"achievable is {P_min:.2e} Pa."
                )
            else:
                try:
                    sol = bracket(_delta_pressure, V_crit - dV, dV, args)
                except Exception:
                    raise Exception(
                        "Cannot find a volume, perhaps you are "
                        "outside of the range of validity for "
                        "the equation of state?"
                    )

        return opt.brentq(_delta_pressure, sol[0], sol[1], args=args)

    def pressure(self, temperature, volume, params):
        """
        Returns the pressure of the mineral at a given temperature and volume
        [Pa]
        """
        debye_T = self._debye_temperature(params["V_0"] / volume, params)
        gr = _grueneisen_parameter_slb(
            params["V_0"], volume, params["grueneisen_0"], params["q_0"]
        )
        # does not depend on pressure
        # thermal energy at temperature T
        E_th = debye.thermal_energy(temperature, debye_T, params["n"])
        # thermal energy at reference temperature
        E_th_ref = debye.thermal_energy(params["T_0"], debye_T, params["n"])

        b_iikk = 9.0 * params["K_0"]  # EQ 28
        b_iikkmm = 27.0 * params["K_0"] * (params["Kprime_0"] - 4.0)  # EQ 29
        f = 0.5 * (pow(params["V_0"] / volume, 2.0 / 3.0) - 1.0)  # EQ 24
        P = (1.0 / 3.0) * (pow(1.0 + 2.0 * f, 5.0 / 2.0)) * (
            (b_iikk * f) + (0.5 * b_iikkmm * pow(f, 2.0))
        ) + gr * (
            E_th - E_th_ref
        ) / volume  # EQ 21
        if self.conductive:
            P += el.pressure(
                temperature,
                volume,
                params["T_0"],
                params["V_0"],
                params["bel_0"],
                params["gel"],
            )
        return P

    def isothermal_bulk_modulus_reuss(self, pressure, temperature, volume, params):
        """
        Returns isothermal bulk modulus :math:`[Pa]`
        """
        T_0 = params["T_0"]
        debye_T = self._debye_temperature(params["V_0"] / volume, params)
        gr = _grueneisen_parameter_slb(
            params["V_0"], volume, params["grueneisen_0"], params["q_0"]
        )

        # thermal energy at temperature T
        E_th = debye.thermal_energy(temperature, debye_T, params["n"])
        # thermal energy at reference temperature
        E_th_ref = debye.thermal_energy(T_0, debye_T, params["n"])

        # heat capacity at temperature T
        C_v = debye.molar_heat_capacity_v(temperature, debye_T, params["n"])
        # heat capacity at reference temperature
        C_v_ref = debye.molar_heat_capacity_v(T_0, debye_T, params["n"])

        q = self.volume_dependent_q(params["V_0"] / volume, params)

        K = (
            bm.bulk_modulus(volume, params)
            + (gr + 1.0 - q) * (gr / volume) * (E_th - E_th_ref)
            - (pow(gr, 2.0) / volume) * (C_v * temperature - C_v_ref * T_0)
        )

        if self.conductive:
            K += volume * el.KToverV(
                temperature,
                volume,
                params["T_0"],
                params["V_0"],
                params["bel_0"],
                params["gel"],
            )
        return K

    def shear_modulus(self, pressure, temperature, volume, params):
        """
        Returns shear modulus. :math:`[Pa]`
        """
        T_0 = params["T_0"]
        debye_T = self._debye_temperature(params["V_0"] / volume, params)
        eta_s = self._isotropic_eta_s(params["V_0"] / volume, params)

        E_th = debye.thermal_energy(temperature, debye_T, params["n"])
        E_th_ref = debye.thermal_energy(T_0, debye_T, params["n"])

        if self.order == 2:
            return (
                bm.shear_modulus_second_order(volume, params)
                - eta_s * (E_th - E_th_ref) / volume
            )
        elif self.order == 3:
            return (
                bm.shear_modulus_third_order(volume, params)
                - eta_s * (E_th - E_th_ref) / volume
            )
        else:
            raise NotImplementedError("")

    def molar_heat_capacity_v(self, pressure, temperature, volume, params):
        """
        Returns heat capacity at constant volume. :math:`[J/K/mol]`
        """
        debye_T = self._debye_temperature(params["V_0"] / volume, params)
        C_v = debye.molar_heat_capacity_v(temperature, debye_T, params["n"])

        if self.conductive:
            C_v += temperature * el.CVoverT(
                volume, params["V_0"], params["bel_0"], params["gel"]
            )
        return C_v

    def thermal_expansivity(self, pressure, temperature, volume, params):
        """
        Returns thermal expansivity. :math:`[1/K]`
        """
        debye_T = self._debye_temperature(params["V_0"] / volume, params)
        C_v = debye.molar_heat_capacity_v(temperature, debye_T, params["n"])
        gr_slb = _grueneisen_parameter_slb(
            params["V_0"], volume, params["grueneisen_0"], params["q_0"]
        )
        K = self.isothermal_bulk_modulus_reuss(pressure, temperature, volume, params)
        alpha = gr_slb * C_v / volume / K

        if self.conductive:
            aKTel = el.aKT(
                temperature, volume, params["V_0"], params["bel_0"], params["gel"]
            )
            alpha += aKTel / K
        return alpha

    def entropy(self, pressure, temperature, volume, params):
        """
        Returns the entropy at the pressure and temperature
        of the mineral [J/K/mol]
        """
        Debye_T = self._debye_temperature(params["V_0"] / volume, params)
        S = debye.entropy(temperature, Debye_T, params["n"])

        if self.conductive:
            S += el.entropy(
                temperature, volume, params["V_0"], params["bel_0"], params["gel"]
            )
        return S

    def helmholtz_free_energy(self, pressure, temperature, volume, params):
        """
        Returns the Helmholtz free energy at the pressure and temperature
        of the mineral [J/mol]
        """
        x = params["V_0"] / volume
        f = 1.0 / 2.0 * (pow(x, 2.0 / 3.0) - 1.0)
        Debye_T = self._debye_temperature(params["V_0"] / volume, params)

        F_quasiharmonic = debye.helmholtz_free_energy(
            temperature, Debye_T, params["n"]
        ) - debye.helmholtz_free_energy(params["T_0"], Debye_T, params["n"])

        b_iikk = 9.0 * params["K_0"]  # EQ 28
        b_iikkmm = 27.0 * params["K_0"] * (params["Kprime_0"] - 4.0)  # EQ 29

        F = (
            params["F_0"]
            + 0.5 * b_iikk * f * f * params["V_0"]
            + (1.0 / 6.0) * params["V_0"] * b_iikkmm * f * f * f
            + F_quasiharmonic
        )

        if self.conductive:
            F += el.helmholtz(
                temperature,
                volume,
                params["T_0"],
                params["V_0"],
                params["bel_0"],
                params["gel"],
            )
        return F

    # Derived properties from here
    # These functions can use pressure as an argument,
    # as pressure should have been determined self-consistently
    # by the point at which these functions are called.
    def grueneisen_parameter(self, pressure, temperature, volume, params):
        """
        Returns grueneisen parameter :math:`[unitless]`
        """
        if self.conductive:
            temperature = max(temperature, 1.0e-6)
            K_T = self.isothermal_bulk_modulus_reuss(
                pressure, temperature, volume, params
            )
            alpha = self.thermal_expansivity(pressure, temperature, volume, params)
            C_v = self.molar_heat_capacity_v(pressure, temperature, volume, params)
            return alpha * K_T * volume / C_v
        else:
            return _grueneisen_parameter_slb(
                params["V_0"], volume, params["grueneisen_0"], params["q_0"]
            )

    def isentropic_bulk_modulus_reuss(self, pressure, temperature, volume, params):
        """
        Returns adiabatic bulk modulus. :math:`[Pa]`
        """
        K_T = self.isothermal_bulk_modulus_reuss(pressure, temperature, volume, params)
        alpha = self.thermal_expansivity(pressure, temperature, volume, params)
        gr = self.grueneisen_parameter(pressure, temperature, volume, params)
        return K_T * (1.0 + gr * alpha * temperature)

    def molar_heat_capacity_p(self, pressure, temperature, volume, params):
        """
        Returns heat capacity at constant pressure. :math:`[J/K/mol]`
        """
        alpha = self.thermal_expansivity(pressure, temperature, volume, params)
        gr = self.grueneisen_parameter(pressure, temperature, volume, params)
        C_v = self.molar_heat_capacity_v(pressure, temperature, volume, params)
        C_p = C_v * (1.0 + gr * alpha * temperature)
        return C_p

    def molar_internal_energy(self, pressure, temperature, volume, params):
        """
        Returns the internal energy at the pressure and temperature
        of the mineral [J/mol]
        """
        return self.helmholtz_free_energy(
            pressure, temperature, volume, params
        ) + temperature * self.entropy(pressure, temperature, volume, params)

    def gibbs_free_energy(self, pressure, temperature, volume, params):
        """
        Returns the Gibbs free energy at the pressure and temperature
        of the mineral [J/mol]
        """
        G = (
            self.helmholtz_free_energy(pressure, temperature, volume, params)
            + pressure * volume
        )
        return G

    def enthalpy(self, pressure, temperature, volume, params):
        """
        Returns the enthalpy at the pressure and temperature
        of the mineral [J/mol]
        """

        return (
            self.helmholtz_free_energy(pressure, temperature, volume, params)
            + temperature * self.entropy(pressure, temperature, volume, params)
            + pressure * volume
        )

    def validate_parameters(self, params):
        """
        Check for existence and validity of the parameters
        """
        if "T_0" not in params:
            params["T_0"] = 300.0

        # If eta_s_0 is not included this is presumably deliberate,
        # as we can model density and bulk modulus just fine without it,
        # so just add it to the dictionary as nan
        # The same goes for the standard state Helmholtz free energy
        if "eta_s_0" not in params:
            params["eta_s_0"] = float("nan")
        if "F_0" not in params:
            params["F_0"] = float("nan")

        # First, let's check the EoS parameters for Tref
        bm.BirchMurnaghanBase.validate_parameters(bm.BirchMurnaghanBase(), params)

        # Now check all the required keys for the
        # thermal part of the EoS are in the dictionary
        expected_keys = ["molar_mass", "n", "Debye_0", "grueneisen_0", "q_0", "eta_s_0"]
        if self.conductive:
            expected_keys.extend(["bel_0", "gel"])

        for k in expected_keys:
            if k not in params:
                raise KeyError("params object missing parameter : " + k)

        # Finally, check that the values are reasonable.
        if params["T_0"] < 0.0:
            warnings.warn("Unusual value for T_0", stacklevel=2)
        if params["molar_mass"] < 0.001 or params["molar_mass"] > 10.0:
            warnings.warn("Unusual value for molar_mass", stacklevel=2)
        if params["n"] < 1.0 or params["n"] > 1000.0:
            warnings.warn("Unusual value for n", stacklevel=2)
        if params["Debye_0"] < 1.0 or params["Debye_0"] > 10000.0:
            warnings.warn("Unusual value for Debye_0", stacklevel=2)
        if params["grueneisen_0"] < -1.0 or params["grueneisen_0"] > 10.0:
            warnings.warn("Unusual value for grueneisen_0", stacklevel=2)
        if params["q_0"] < -20.0 or params["q_0"] > 20.0:
            warnings.warn("Unusual value for q_0", stacklevel=2)
        if params["eta_s_0"] < -10.0 or params["eta_s_0"] > 10.0:
            warnings.warn("Unusual value for eta_s_0", stacklevel=2)


class SLB3(SLBBase):
    """
    SLB equation of state with third order finite strain expansion for the
    shear modulus (this should be preferred, as it is more thermodynamically
    consistent).
    """

    def __init__(self):
        self.order = 3
        self.conductive = False


class SLB2(SLBBase):
    """
    SLB equation of state with second order finite strain expansion for the
    shear modulus.  In general, this should not be used, but sometimes
    shear modulus data is fit to a second order equation of state.  In that
    case, you should use this.  The moral is, be careful!
    """

    def __init__(self):
        self.order = 2
        self.conductive = False


class SLB3Conductive(SLBBase):
    """
    SLB equation of state with third order finite strain expansion for the
    shear modulus and a contribution to the Helmholtz free energy
    that arises from the thermal excitation of electrons (Bukowinski, 1977).
    """

    def __init__(self):
        self.order = 3
        self.conductive = True
